__all__ = [
    "tran_angles",
    "tf_angles",
    "tran_bootstrap",
    "tf_bootstrap",
    "tran_copula_corr",
    "tf_copula_corr",
    "tran_kfolds",
    "tf_kfolds",
    "tran_md",
    "tf_md",
]

from grama import add_pipe, pipe, copy_meta, Intention, mse, rsq
from grama import (
    tf_bind_cols,
    tf_filter,
    tf_summarize,
    tf_drop,
    tf_mutate,
    var_in,
    ev_df,
)
from collections import ChainMap
from numbers import Integral
from numpy import arange, ceil, zeros, std, quantile, nan, triu_indices, unique
from numpy.random import choice, permutation
from numpy.random import seed as set_seed
from pandas import concat, DataFrame, melt
from pandas.api.types import is_numeric_dtype
from scipy.linalg import subspace_angles
from scipy.stats import norm
from .string_helpers import str_detect, str_replace
from toolz import curry

X = Intention()


## k-Fold CV utility
# --------------------------------------------------
@curry
def tran_kfolds(
    df,
    k=None,
    ft=None,
    out=None,
    var_fold=None,
    suffix="_mean",
    summaries=None,
    tf=tf_summarize,
    shuffle=True,
    seed=None,
):
    r"""Perform k-fold CV

    Perform k-fold cross-validation (CV) using a given fitting procedure (ft).
    Optionally provide a fold identifier column, or (randomly) assign folds.

    Args:
        df (DataFrame): Data to pass to given fitting procedure
        ft (gr.ft_): Partially-evaluated grama fit function; defines model fitting
            procedure and outputs to aggregate
        tf (gr.tf_): Partially-evaluated grama transform function; evaluation of
            fitted model will be passed to tf and provided with keyword arguments
            from summaries
        out (list or None): Outputs for which to compute `summaries`; None uses ft.out
        var_fold (str or None): Column to treat as fold identifier; overrides `k`
        suffix (str): Suffix for predicted value; used to distinguish between predicted and actual
        summaries (dict of functions): Summary functions to pass to tf; will be evaluated
            for outputs of ft. Each summary must have signature summary(f_pred, f_meas).
            Grama includes builtin options: gr.mse, gr.rmse, gr.rel_mse, gr.rsq, gr.ndme
        k (int): Number of folds; k=5 to k=10 recommended [1]
        shuffle (bool): Shuffle the data before CV? True recommended [1]

    Notes:
        - Many grama functions support *partial evaluation*; this allows one to specify things like hyperparameters in fitting functions without providing data and executing the fit. You can take advantage of this functionality to easly do hyperparameter studies.

    Returns:
        DataFrame: Aggregated results within each of k-folds using given model and
            summary transform

    References:
        [1] James, Witten, Hastie, and Tibshirani, "An introduction to statistical learning" (2017), Chapter 5. Resampling Methods

    Examples:

        >>> import grama as gr
        >>> from grama.data import df_stang
        >>> from grama.fit import ft_rf
        >>> df_kfolds = (
        >>>     df_stang
        >>>     >> gr.tf_kfolds(
        >>>         k=5,
        >>>         ft=ft_rf(out=["thick"], var=["E", "mu"]),
        >>>     )

    """
    ## Check invariants
    if ft is None:
        raise ValueError("Must provide ft keyword argument")
    if (k is None) and (var_fold is None):
        print("... tran_kfolds is using default k=5")
        k = 5
    if summaries is None:
        print("... tran_kfolds is using default summaries mse and rsq")
        summaries = dict(mse=mse, rsq=rsq)

    n = df.shape[0]
    ## Handle custom folds
    if not (var_fold is None):
        ## Check for a valid var_fold
        if not (var_fold in df.columns):
            raise ValueError("var_fold must be in df.columns or None")
        ## Build folds
        levels = unique(df[var_fold])
        k = len(levels)
        print("... tran_kfolds found {} levels via var_folds".format(k))
        Is = []
        for l in levels:
            Is.append(list(arange(n)[df[var_fold] == l]))

    else:
        ## Shuffle data indices
        if shuffle:
            if seed:
                set_seed(seed)
            I = permutation(n)
        else:
            I = arange(n)
        ## Build folds
        di = int(ceil(n / k))
        Is = [I[i * di : min((i + 1) * di, n)] for i in range(k)]

    ## Iterate over folds
    df_res = DataFrame()
    for i in range(k):
        ## Train by out-of-fold data
        md_fit = df >> tf_filter(~var_in(X.index, Is[i])) >> ft

        ## Determine predicted and actual
        if out is None:
            out = str_replace(md_fit.out, suffix, "")
        else:
            out = str_replace(out, suffix, "")

        ## Test by in-fold data
        df_pred = md_fit >> ev_df(
            df=df >> tf_filter(var_in(X.index, Is[i])), append=False
        )

        ## Specialize summaries for output names
        summaries_all = ChainMap(
            *[
                {
                    key + "_" + o: fun(X[o + suffix], X[o])
                    for key, fun in summaries.items()
                }
                for o in out
            ]
        )

        ## Aggregate
        df_summary_tmp = (
            df_pred
            >> tf_bind_cols(df[out] >> tf_filter(var_in(X.index, Is[i])))
            >> tf(**summaries_all)
            # >> tf_mutate(_kfold=i)
        )

        if var_fold is None:
            df_summary_tmp = df_summary_tmp >> tf_mutate(_kfold=i)
        else:
            df_summary_tmp[var_fold] = levels[i]

        df_res = concat((df_res, df_summary_tmp), axis=0).reset_index(drop=True)

    return df_res


tf_kfolds = add_pipe(tran_kfolds)

## Bootstrap utility
# --------------------------------------------------
@curry
def tran_bootstrap(
    df, tran=None, n_boot=500, n_sub=25, con=0.90, col_sel=None, seed=None
):
    r"""Estimate bootstrap confidence intervals

    Estimate bootstrap confidence intervals for a given transform. Uses the
    "bootstrap-t" procedure discussed in Efron and Tibshirani (1993).

    Args:
        df (DataFrame): Data to bootstrap
        tran (grama tran_ function): Transform procedure which generates statistic
        n_boot (numeric): Monte Carlo resamples for bootstrap
        n_sub (numeric): Nested resamples to estimate SE
        con (float): Confidence level
        col_sel (list(string)): Columns to include in bootstrap calculation

    Returns:
        DataFrame: Results of tran(df), plus _lo and _up columns for
        numeric columns

    References and notes:
       Efron and Tibshirani (1993) "The bootstrap-t procedure... is
       particularly applicable to location statistics like the sample mean....
       The bootstrap-t method, at least in its simple form, cannot be trusted
       for more general problems, like setting a confidence interval for a
       correlation coefficient."

    Examples:

    """
    ## Set seed only if given
    if seed is not None:
        set_seed(seed)

    ## Ensure sample count is int
    if not isinstance(n_boot, Integral):
        print("tran_bootstrap() is rounding n_boot...")
        n_boot = int(n_boot)
    if not isinstance(n_sub, Integral):
        print("tran_bootstrap() is rounding n_sub...")
        n_sub = int(n_sub)

    ## Base results
    df_base = tran(df)

    ## Select columns for bootstrap
    col_numeric = list(df_base.select_dtypes(include="number").columns)
    if not (col_sel is None):
        col_numeric = list(set(col_numeric).intersection(set(col_sel)))

    ## Setup
    n_samples = df.shape[0]
    n_row = df_base.shape[0]
    n_col = len(col_numeric)
    alpha = (1 - con) / 2
    theta_hat = df_base[col_numeric].values

    theta_all = zeros((n_boot, n_row, n_col))
    se_boot_all = zeros((n_boot, n_row, n_col))
    z_all = zeros((n_boot, n_row, n_col))
    theta_sub = zeros((n_sub, n_row, n_col))

    ## Main loop
    for ind in range(n_boot):
        ## Construct resample
        Ib = choice(n_samples, size=n_samples, replace=True)
        df_tmp = copy_meta(df, df.iloc[Ib,])
        theta_all[ind] = tran(df_tmp)[col_numeric].values

        ## Internal loop to approximate SE
        for jnd in range(n_sub):
            Isub = Ib[choice(n_samples, size=n_samples, replace=True)]
            df_tmp = copy_meta(df, df.iloc[Isub,])
            theta_sub[jnd] = tran(df_tmp)[col_numeric].values
        se_boot_all[ind] = std(theta_sub, axis=0)

        ## Construct approximate pivot
        z_all[ind] = (theta_all[ind] - theta_hat) / se_boot_all[ind]

    ## Compute bootstrap table
    t_lo, t_hi = quantile(z_all, q=[1 - alpha, alpha], axis=0)

    ## Estimate bootstrap intervals
    se = std(theta_all, axis=0)
    theta_lo = theta_hat - t_lo * se
    theta_hi = theta_hat - t_hi * se

    ## Assemble output data
    col_lo = list(map(lambda s: s + "_lo", col_numeric))
    col_hi = list(map(lambda s: s + "_up", col_numeric))

    df_lo = DataFrame(data=theta_lo, columns=col_lo)
    df_hi = DataFrame(data=theta_hi, columns=col_hi)

    df_ci = concat((df_lo, df_hi), axis=1).sort_index(axis=1)
    df_ci.index = df_base.index

    return concat((df_base, df_ci), axis=1)


tf_bootstrap = add_pipe(tran_bootstrap)


## Assess subspace angles
# --------------------------------------------------
def tran_angles(df, df2):
    r"""Subspace angles

    Compute the subspace angles between two matrices. A wrapper for
    scipy.linalg.subspace_angles that corrects for column ordering. Row ordering
    is assumed.

    Args:
        df (DataFrame): First matrix to compare
        df2 (DataFrame): Second matrix to compare

    Returns:
        array: Array of angles (in radians)

    Examples:

        >>> import grama as gr
        >>> import pandas as pd
        >>> df = pd.DataFrame(dict(v=[+1, +1]))
        >>> df_v1 = pd.DataFrame(dict(w=[+1, -1]))
        >>> df_v2 = pd.DataFrame(dict(w=[+1, +1]))
        >>> theta1 = angles(df, df_v1)
        >>> theta2 = angles(df, df_v2)

    """
    ## Compute subspace angles
    A1 = df.values
    A2 = df2.values

    return subspace_angles(A1, A2)


tf_angles = add_pipe(tran_angles)


## Compute Gaussian copula correlations from data
# --------------------------------------------------
def tran_copula_corr(df, model=None, density=None):
    r"""Compute Gaussian copula correlations from data

    Convenience function to fit a Gaussian copula (correlations) based on data
    and pre-fitted marginals. Intended for use with gr.comp_copula_gaussian().
    Must provide either `model` or `density`.

    Args:
        df (DataFrame): Matrix of data for correlation estimation
        model (gr.Model): Model with defined marginals
        density (gr.Density): Density with defined marginals

    Returns:
        DataFrame: Correlation data ready for use with gr.comp_copula_gaussian()

    Examples:

        >>> import grama as gr
        >>> from grama.data import df_stang
        >>> md = gr.Model() >> \
        >>>     gr.cp_marginals(
        >>>         E=gr.marg_named(df_stang.E, "norm"),
        >>>         mu=gr.marg_named(df_stang.mu, "beta"),
        >>>         thick=gr.marg_named(df_stang.thick, "norm")
        >>>     )
        >>> df_corr = gr.tran_copula_corr(df_stang, model=md)

    """
    if density is None:
        density = model.density

    ## Check invariants
    if not set(density.marginals.keys()).issubset(set(df.columns)):
        raise ValueError("df must have columns for all var_rand")

    ## Convert data
    df_res = density.sample2pr(df)
    df_norm = df_res.apply(norm.ppf)

    ## Compute correlations
    df_mat = df_norm.corr()
    Ind = triu_indices(len(density.marginals), 1)

    ## Arrange
    var_rand = df_mat.columns
    var1_all = []
    var2_all = []
    corr_all = []

    for i, j in zip(Ind[0], Ind[1]):
        var1_all.append(var_rand[i])
        var2_all.append(var_rand[j])
        corr_all.append(df_mat.iloc[i, j])

    return DataFrame(dict(var1=var1_all, var2=var2_all, corr=corr_all))


tf_copula_corr = add_pipe(tran_copula_corr)

## Model as transform
# --------------------------------------------------
@curry
def tran_md(df, md=None, append=True):
    r"""Model as transform

    Use a model to transform data; useful when pre-processing data to evaluate a
    model.

    Args:
        df (DataFrame): Data to merge
        md (gr.Model): Model to use as transform

    Returns:
        DataFrame: Output of evaluated model

    Examples:
        >>> import grama as gr
        >>> from grama.models import make_cantilever_beam
        >>> md_beam = make_cantilever_beam()
        >>> df_res = (
        >>>     md_beam
        >>>     >> gr.ev_monte_carlo(n=1e3, df_det="nom", skip=True, seed=101)
        >>>     >> gr.tf_sp(n=100)
        >>>     >> gr.tf_md(md=md_beam)
        >>> )

    """
    if md is None:
        raise ValueError("No input md given")
    if len(md.functions) == 0:
        raise ValueError("Given model has no functions")
    out_intersect = set(df.columns).intersection(md.out)
    if len(out_intersect) > 0:
        print(
            "... provided columns intersect model output.\n"
            + "tran_md() is dropping {}".format(out_intersect)
        )

    df_res = md.evaluate_df(df)

    if append:
        df_res = concat(
            [df.reset_index(drop=True).drop(md.out, axis=1, errors="ignore"), df_res,],
            axis=1,
        )

    return df_res


tf_md = add_pipe(tran_md)



__all__ = ["comp_metamodel", "cp_metamodel"]

## Fitting via statsmodels package
from grama import add_pipe, pipe
from grama.eval import eval_lhs
from grama.fit import fit_ols
from toolz import curry


## Fit a metamodel
# --------------------------------------------------
@curry
def comp_metamodel(model, n=1, ev=None, ft=None, seed=None):
    r"""Create a metamodel

    Composition: Create a metamodel from an existing model. This convenience
    function essentially applies a recipe of Evaluation followed by Fitting.
    Default methods are Latin Hypercube Evaluation and Ordinary Least Squares
    Fitting with linear features.

    Args:
        model (gr.model): Original model, to be evaluated and fit
        n (numeric): Number of samples to draw
        ev (gr.eval_): Evaluation strategy, default eval_lhs
        ft (gr.fit_): Fitting strategy, default fit_ols w/ linear features
        seed (int): Random seed, default None

    Returns:
        gr.model: Metamodel

    """
    ## Extract model information
    inputs = model.domain.inputs
    outputs = model.outputs

    ## Assign default arguments
    if ev is None:
        ev = eval_lhs

    if ft is None:
        # Linear features for each output
        sum_inputs = "+".join(inputs)
        formulae = list(map(lambda output: output + "~" + sum_inputs, outputs))

        ft = lambda df: fit_ols(
            df, formulae=formulae, domain=model.domain, density=model.density
        )

    ## Generate data
    df_results = ev(model, n_samples=n, seed=seed)

    ## Fit a model
    model = ft(df_results)

    return model


cp_metamodel = add_pipe(comp_metamodel)



__all__ = [
    "df_diamonds",
    "df_ruff",
    "df_stang",
    "df_stang_wide",
    "df_trajectory_full",
    "df_trajectory_windowed",
    "df_shewhart",
]

import os
from pandas import read_csv
from pathlib import Path


path_this = Path(__file__)
path_grama = path_this.parents[1]

# Stang (wdie form)
df_stang_wide = read_csv(Path(path_grama / "data/stang.csv"))

# Stang (tidy form)
df_stang = read_csv(Path(path_grama / "data/stang_long.csv"))

# Stang (tidy form)
df_diamonds = read_csv(Path(path_grama / "data/diamonds.csv"))

# Ruff (tidy form)
df_ruff = read_csv(Path(path_grama / "data/ruff.csv"))

# Trajectories
df_trajectory_full = read_csv(Path(path_grama / "data/trajectory_full.csv"))
df_trajectory_windowed = read_csv(Path(path_grama / "data/trajectory_windowed.csv"))

# Shewhart
df_shewhart = read_csv(Path(path_grama / "data/shewhart1931-table3.csv"))



__all__ = [
    "tran_polyridge",
    "tf_polyridge",
]

## Transforms via psdr package
try:
    from psdr import PolynomialRidgeApproximation

except ModuleNotFoundError:
    pass

from grama import add_pipe
from numpy import number as npnumber
from pandas import DataFrame
from toolz import curry

## Polynomial Ridge Approximation
# --------------------------------------------------
@curry
def tran_polyridge(
    df, var=None, out=None, n_dim=None, n_degree=None, **kwargs
):
    r"""Polynomial Ridge Approximation for parameter space reduction

    Apply the polynomial ridge approximation to seek a low-dimensional subspace
    among inputs that best explain variability in a selected output.

    Args:
        df (DataFrame): Target dataset to reduce
        var (list or None): Variables in df on which to perform dimension reduction.
            Use None to compute with all numeric variables, except `out`.
        out (string): Name of target output
        n_dim (int): Target dimensionality
        n_degree (int): Fitted polynomial (total) order

    Kwargs:
        n_init (int): Number of iterations in optimization

    Returns:
        DataFrame: Set of weights defining the dimension-reduced space.

    Notes:
        - A wrapper for psdr.PolynomialRidgeApproximation

    References:
        - J. M. Hokanson and Paul G. Constantine. Data-driven Polynomial Ridge Approximation Using Variable Projection. SIAM J. Sci. Comput. Vol 40, No 3, pp A1566–A1589, DOI:10.1137/17M1117690.

    Examples:
        >>> import grama as gr
        >>> from plotnine import *
        >>> from grama.psdr import tf_polyridge
        >>> DF = gr.Intention()
        >>>
        >>> ## Set up a dataset to reduce
        >>> df_data = (
        >>>     gr.df_make(x=range(10))
        >>>     >> gr.tf_outer(gr.df_make(y=range(10)))
        >>>     >> gr.tf_outer(gr.df_make(z=range(10)))
        >>>     >> gr.tf_mutate(f=DF.x - DF.y)
        >>> )
        >>>
        >>> ## Use polynomial ridge approximation to derive weights
        >>> df_weights = (
        >>>     df_data
        >>>     >> tf_polyridge(out="f", n_dim=1, n_degree=1)
        >>> )
        >>>
        >>> ## Construct a shadow plot
        >>> (
        >>>     df_data
        >>>     >> gr.tf_inner(df_weights)
        >>>     >> ggplot(aes("dot", "f"))
        >>>     + geom_point()
        >>> )

    """
    ## Check invariants
    if out is None:
        raise ValueError("Must select an `out` column.")
    if not (out in df.columns):
        raise ValueError("out={} is not a column in provided dataframe.".format(out))

    if var is None:
        var = set(df.select_dtypes(include=[npnumber]).columns.values)
        var = list(var.difference({out}))
    else:
        var = list(var).copy()
        diff = set(var).difference(set(df.columns))
        if len(diff) > 0:
            raise ValueError(
                "`var` must be subset of `df.columns`\n" "diff = {}".format(diff)
            )

    if (n_degree == 1) and (n_dim > 1):
        raise ValueError(
            "n_dim > 1 cannot work with n_degree == 1; this would be " +
            "an over-parameterized linear model. Try again with n_degree > 1."
        )

    ## Compute subspace reduction
    try:
        pr = PolynomialRidgeApproximation(
            n_degree,
            n_dim,
            **kwargs,
        )
        pr.fit(df[var].values, df[out].values)

    except NameError as e:
        error_string = str(e)
        raise NameError(
            error_string +
            "\n\nThis function requires the `psdr` package. " +
            "Try running the following to install the package:\n"
            "    pip install psdr"
        )

    ## Package the results
    df_res = DataFrame(
        data=pr.U.T,
        columns=var
    )
    return df_res

tf_polyridge = add_pipe(tran_polyridge)



__all__ = [
    "eval_form_pma",
    "ev_form_pma",
    "eval_form_ria",
    "ev_form_ria",
]

from grama import add_pipe, pipe, custom_formatwarning
from .eval_defaults import eval_df
from numpy import array, argmin, ones, eye, zeros, sqrt, NaN, max
from numpy.linalg import norm as length
from numpy.random import multivariate_normal
from pandas import DataFrame, concat
from scipy.optimize import minimize
from toolz import curry


## Utility Functions
# --------------------------------------------------
def make_T(model, df_corr):
    """Make covariance matrix from list

    Construct a covariance matrix from given entries. Check for correctness
    against a model.

    """
    ## TODO
    raise NotImplementedError


## FORM
# --------------------------------------------------
@curry
def eval_form_pma(
    model,
    betas=None,
    cons=None,
    df_corr=None,
    df_det=None,
    append=True,
    tol=1e-3,
    n_maxiter=25,
    n_restart=1,
    verbose=False,
):
    r"""Tail quantile via FORM PMA

    Approximate the desired tail quantiles using the performance measure
    approach (PMA) of the first-order reliability method (FORM) [1]. Select
    limit states to minimize at desired quantile with `betas`. Provide
    confidence levels `cons` and estimator covariance `df_corr` to compute with
    margin in beta [2].

    Args:
        model (gr.Model): Model to analyze
        betas (dict): Target reliability indices;
            key   = limit state name; must be in model.out
            value = reliability index; beta = Phi^{-1}(reliability)
        cons (dict or None): Target confidence levels;
            key   = limit state name; must be in model.out
            value = confidence level, \in (0, 1)
        df_corr (DataFrame or None): Sampling distribution covariance entries;
            parameters with no information assumed to be known exactly.
        df_det (DataFrame): Deterministic levels for evaluation; use "nom"
            for nominal deterministic levels.
        n_maxiter (int): Maximum iterations for each optimization run
        n_restart (int): Number of restarts (== number of optimization runs)
        append (bool): Append MPP results for random values?
        verbose (bool): Print optimization results?

    Returns:
        DataFrame: Results of MPP search

    Notes:
        - Since FORM PMA relies on optimization over the limit state, it is
          often beneficial to scale your limit state to keep values near unity.

    References:
        - [1] Tu, Choi, and Park, "A new study on reliability-based design optimization," Journal of Mechanical Design, 1999
        - [2] del Rosario, Fenrich, and Iaccarino, "Fast precision margin with the first-order reliability method," AIAA Journal, 2019

    """
    ## Check invariants
    if betas is None:
        raise ValueError(
            "Must provide `betas` keyword argument to define reliability targets"
        )
    if not set(betas.keys()).issubset(set(model.out)):
        raise ValueError("betas.keys() must be subset of model.out")
    if not cons is None:
        if not (set(cons.keys()) == set(betas.keys())):
            raise ValueError("cons.keys() must be same as betas.keys()")
        else:
            if df_corr is None:
                raise ValueError("Must provide df_corr is using cons")
        raise NotImplementedError

    df_det = model.var_outer(
        DataFrame(data=zeros((1, model.n_var_rand)), columns=model.var_rand),
        df_det=df_det,
    )
    df_det = df_det[model.var_det]

    df_return = DataFrame()
    for ind in range(df_det.shape[0]):
        ## Loop over objectives
        for key in betas.keys():
            ## Temp dataframe
            df_inner = df_det.iloc[[ind]].reset_index(drop=True)

            ## Construct lambdas
            def objective(z):
                ## Transform: standard normal-to-random variable
                df_norm = DataFrame(data=[z], columns=model.var_rand)
                df_rand = model.norm2rand(df_norm)
                df = model.var_outer(df_rand, df_det=df_inner)

                df_res = eval_df(model, df=df)
                g = df_res[key].iloc[0]

                # return (g, jac)
                return g

            def con_beta(z):
                return z.dot(z) - (betas[key]) ** 2

            ## Use conservative direction for initial guess
            signs = array([model.density.marginals[k].sign for k in model.var_rand])
            if length(signs) > 0:
                z0 = betas[key] * signs / length(signs)
            else:
                z0 = betas[key] * ones(model.n_var_rand) / sqrt(model.n_var_rand)

            ## Minimize
            res_all = []
            for jnd in range(n_restart):
                res = minimize(
                    objective,
                    z0,
                    args=(),
                    method="SLSQP",
                    jac=False,
                    tol=tol,
                    options={"maxiter": n_maxiter, "disp": False},
                    constraints=[{"type": "eq", "fun": con_beta}],
                )
                # Append only a successful result
                if res["status"] == 0:
                    res_all.append(res)
                # Set a random start; repeat
                z0 = multivariate_normal([0] * model.n_var_rand, eye(model.n_var_rand))
                z0 = z0 / length(z0) * betas[key]

            # Choose value among restarts
            n_iter_total = sum([res_all[i].nit for i in range(len(res_all))])
            if len(res_all) > 0:
                i_star = argmin([res.fun for res in res_all])
                x_star = res_all[i_star].x
                fun_star = res_all[i_star].fun
                if verbose:
                    print("out = {}: Optimization successful".format(key))
                    print("n_iter = {}".format(res_all[i_star].nit))
                    print("n_iter_total = {}".format(n_iter_total))
            else:
                ## WARNING
                x_star = [NaN] * model.n_var_rand
                fun_star = NaN
                if verbose:
                    print("out = {}: Optimization unsuccessful".format(key))
                    print("n_iter = {}".format(res_all[i_star].nit))
                    print("n_iter_total = {}".format(n_iter_total))

            ## Extract results
            if append:
                df_inner = concat(
                    (
                        df_inner,
                        model.norm2rand(
                            DataFrame(data=[x_star], columns=model.var_rand)
                        ),
                    ),
                    axis=1,
                    sort=False,
                )
            df_inner[key] = [fun_star]
            df_return = concat((df_return, df_inner), axis=0, sort=False)

    if not append:
        df_return = (
            df_return.groupby(model.var_det)
            .agg({s: max for s in betas.keys()})
            .reset_index()
        )

    return df_return


ev_form_pma = add_pipe(eval_form_pma)


@curry
def eval_form_ria(
    model,
    limits=None,
    cons=None,
    df_corr=None,
    df_det=None,
    append=True,
    tol=1e-3,
    n_maxiter=25,
    n_restart=1,
    verbose=False,
):
    r"""Tail reliability via FORM RIA

    Approximate the desired tail probability using the reliability index
    approach (RIA) of the first-order reliability method (FORM) [1]. Select
    limit states to analyze with list input `limits`. Provide confidence levels
    `cons` and estimator covariance `df_corr` to compute with margin in beta
    [2].

    Args:
        model (gr.Model): Model to analyze
        limits (list): Target limit states; must be in model.out; limit state
            assumed to be critical at g == 0
        cons (dict or None): Target confidence levels;
            key   = limit state name; must be in model.out
            value = confidence level, \in (0, 1)
        df_corr (DataFrame or None): Sampling distribution covariance entries;
            parameters with no information assumed to be known exactly.
        df_det (DataFrame): Deterministic levels for evaluation; use "nom"
            for nominal deterministic levels.
        n_maxiter (int): Maximum iterations for each optimization run
        n_restart (int): Number of restarts (== number of optimization runs)
        append (bool): Append MPP results for random values?
        verbose (bool): Print optimization results?

    Returns:
        DataFrame: Results of MPP search

    Notes:
        - Since FORM RIA relies on optimization over the limit state, it is
          often beneficial to scale your limit state to keep values near unity.

    References:
        - [1] Tu, Choi, and Park, "A new study on reliability-based design optimization," Journal of Mechanical Design, 1999
        - [2] del Rosario, Fenrich, and Iaccarino, "Fast precision margin with the first-order reliability method," AIAA Journal, 2019

    """
    ## Check invariants
    if limits is None:
        raise ValueError(
            "Must provide `limits` keyword argument to define reliability targets"
        )
    if not set(limits).issubset(set(model.out)):
        raise ValueError("`limits` must be subset of model.out")
    if not cons is None:
        if not (set(cons.keys()) == set(limits)):
            raise ValueError("cons.keys() must be same as limits")
        else:
            if df_corr is None:
                raise ValueError("Must provide df_corr is using cons")
        raise NotImplementedError

    df_det = model.var_outer(
        DataFrame(data=zeros((1, model.n_var_rand)), columns=model.var_rand),
        df_det=df_det,
    )
    df_det = df_det[model.var_det]

    # df_return = DataFrame(columns=model.var_rand + model.var_det + limits)
    df_return = DataFrame()
    for ind in range(df_det.shape[0]):
        ## Loop over objectives
        for key in limits:
            ## Temp dataframe
            df_inner = df_det.iloc[[ind]].reset_index(drop=True)

            ## Construct lambdas
            def fun_jac(z):
                ## Squared reliability index
                fun = z.dot(z)
                jac = 2 * z * length(z)

                return (fun, jac)

            def con_limit(z):
                ## Transform: standard normal-to-random variable
                df_norm = DataFrame(data=[z], columns=model.var_rand)
                df_rand = model.norm2rand(df_norm)
                df = model.var_outer(df_rand, df_det=df_inner)

                ## Eval limit state
                df_res = eval_df(model, df=df)
                g = df_res[key].iloc[0]

                return g

            ## Use conservative direction for initial guess
            signs = array([model.density.marginals[k].sign for k in model.var_rand])
            if length(signs) > 0:
                z0 = signs / length(signs)
            else:
                z0 = ones(model.n_var_rand) / sqrt(model.n_var_rand)

            ## Minimize
            res_all = []
            for jnd in range(n_restart):
                res = minimize(
                    fun_jac,
                    z0,
                    args=(),
                    method="SLSQP",
                    jac=True,
                    tol=tol,
                    options={"maxiter": n_maxiter, "disp": False},
                    constraints=[{"type": "eq", "fun": con_limit}],
                )
                # Append only a successful result
                if res["status"] == 0:
                    res_all.append(res)
                # Set a random start; repeat
                z0 = multivariate_normal([0] * model.n_var_rand, eye(model.n_var_rand))
                z0 = z0 / length(z0)

            # Choose value among restarts
            n_iter_total = sum([res_all[i].nit for i in range(len(res_all))])
            if len(res_all) > 0:
                i_star = argmin([res.fun for res in res_all])
                x_star = res_all[i_star].x
                fun_star = sqrt(res_all[i_star].fun)
                if verbose:
                    print("out = {}: Optimization successful".format(key))
                    print("n_iter = {}".format(res_all[i_star].nit))
                    print("n_iter_total = {}".format(n_iter_total))
            else:
                ## WARNING
                x_star = [NaN] * model.n_var_rand
                fun_star = NaN
                if verbose:
                    print("out = {}: Optimization unsuccessful".format(key))
                    print("n_iter = {}".format(res_all[i_star].nit))
                    print("n_iter_total = {}".format(n_iter_total))

            ## Extract results
            if append:
                df_inner = concat(
                    (
                        df_inner,
                        model.norm2rand(
                            DataFrame(data=[x_star], columns=model.var_rand)
                        ),
                    ),
                    axis=1,
                    sort=False,
                )
            df_inner["beta_" + key] = [fun_star]
            df_return = concat((df_return, df_inner), axis=0, sort=False)

    if not append:
        df_return = (
            df_return.groupby(model.var_det) \
                     .agg({"beta_" + s: max for s in limits}).reset_index()
        )

    return df_return


ev_form_ria = add_pipe(eval_form_ria)



__all__ = [
    "eval_sample",
    "ev_sample",
    "eval_sinews",
    "ev_sinews",
    "eval_hybrid",
    "ev_hybrid",
]

import warnings
import grama as gr
from grama import add_pipe, CopulaIndependence, custom_formatwarning, eval_df, pipe
from numbers import Integral
from numpy import tile, linspace, zeros, isfinite
from numpy.linalg import cholesky, inv
from numpy.random import random
from numpy.random import seed as set_seed
from pandas import DataFrame
from scipy.stats import norm, lognorm
from toolz import curry


warnings.formatwarning = custom_formatwarning


## Random sampling
# --------------------------------------------------
@curry
def eval_sample(model, n=None, df_det=None, seed=None, append=True, skip=False):
    r"""Draw a random sample

    Evaluates a given model at a given DataFrame. Generates outer product
    with deterministic samples.

    Args:
        model (gr.Model): Model to evaluate
        n (numeric): number of observations to draw
        df_det (DataFrame): Deterministic levels for evaluation; use "nom"
            for nominal deterministic levels.
        seed (int): random seed to use
        append (bool): Append results to random values?
        skip (bool): Skip evaluation of the functions?

    Returns:
        DataFrame: Results of evaluation or unevaluated design

    Examples:

        >>> import grama as gr
        >>> from grama.models import make_test
        >>> md = make_test()
        >>> df = md >> gr.ev_sample(n=1e2, df_det="nom")
        >>> df.describe()

    """
    ## Check invariants
    if n is None:
        raise ValueError("Must provide a valid n value.")

    ## Set seed only if given
    if seed is not None:
        set_seed(seed)

    ## Ensure sample count is int
    if not isinstance(n, Integral):
        print("eval_sample() is rounding n...")
        n = int(n)

    ## Draw samples
    df_rand = model.density.sample(n=n, seed=seed)
    ## Construct outer-product DOE
    df_samp = model.var_outer(df_rand, df_det=df_det)

    if skip:
        ## Evaluation estimate
        runtime_est = model.runtime(df_samp.shape[0])
        if runtime_est > 0:
            print(
                "Estimated runtime for design with model ({0:1}):\n  {1:4.3} sec".format(
                    model.name, runtime_est
                )
            )
        else:
            print("Design runtime estimates unavailable; model has no timing data.")

        ## Attach metadata
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            df_samp._plot_info = {
                "type": "sample_inputs",
                "var": model.var_rand,
            }

        return df_samp

    df_res = eval_df(model, df=df_samp, append=append)
    ## Attach metadata
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        df_res._plot_info = {"type": "sample_outputs", "out": model.out}

    return df_res


ev_sample = add_pipe(eval_sample)

## Marginal sweeps with random origins
# --------------------------------------------------
@curry
def eval_sinews(
    model,
    n_density=10,
    n_sweeps=3,
    seed=None,
    df_det=None,
    varname="sweep_var",
    indname="sweep_ind",
    append=True,
    skip=False,
):
    r"""Sweep study

    Perform coordinate sweeps over each model random variable ("sinew" design).
    Use random starting points drawn from the joint density.

    Use gr.plot_auto() to construct a quick visualization of the output
    dataframe. Use `skip` version to visualize the design, and non-skipped
    version to visualize the results.

    Args:
        model (gr.Model): Model to evaluate
        n_density (numeric): Number of points along each sweep
        n_sweeps (numeric): Number of sweeps per-random variable
        seed (int): Random seed to use
        df_det (DataFrame): Deterministic levels for evaluation;
            use "nom" for nominal deterministic levels,
            use "swp" to sweep deterministic variables
        varname (str): Column name to give for sweep variable; default="sweep_var"
        indname (str): Column name to give for sweep index; default="sweep_ind"
        append (bool): Append results to conservative inputs?
        skip (bool): Skip evaluation of the functions?

    Returns:
        DataFrame: Results of evaluation or unevaluated design

    Examples:

        >>> import grama as gr
        >>> md = gr.make_cantilever_beam()
        >>> # Skip evaluation, vis. design
        >>> df_design = md >> gr.ev_sinews(df_det="nom", skip=True)
        >>> df_design >> gr.pt_auto()
        >>> # Vis results
        >>> df_sinew = md >> gr.ev_sinews(df_det="nom")
        >>> df_sinew >> gr.pt_auto()

    """
    ## Override model if deterministic sweeps desired
    if df_det == "swp":
        ## Collect sweep-able deterministic variables
        var_sweep = list(
            filter(
                lambda v: isfinite(model.domain.get_width(v))
                & (model.domain.get_width(v) > 0),
                model.var_det,
            )
        )
        ## Generate pseudo-marginals
        dicts_var = {}
        for v in var_sweep:
            dicts_var[v] = {
                "dist": "uniform",
                "loc": model.domain.get_bound(v)[0],
                "scale": model.domain.get_width(v),
            }
        ## Overwrite model
        model = gr.comp_marginals(model, **dicts_var)
        ## Restore flag
        df_det = "nom"

    ## Set seed only if given
    if seed is not None:
        set_seed(seed)

    ## Ensure sample count is int
    if not isinstance(n_density, Integral):
        print("eval_sinews() is rounding n_density...")
        n_density = int(n_density)
    if not isinstance(n_sweeps, Integral):
        print("eval_sinews() is rounding n_sweeps...")
        n_sweeps = int(n_sweeps)

    ## Build quantile sweep data
    q_random = tile(random((1, model.n_var_rand, n_sweeps)), (n_density, 1, 1))
    q_dense = linspace(0, 1, num=n_density)
    Q_all = zeros((n_density * n_sweeps * model.n_var_rand, model.n_var_rand))
    C_var = ["tmp"] * (n_density * n_sweeps * model.n_var_rand)
    C_ind = [0] * (n_density * n_sweeps * model.n_var_rand)

    ## Interlace
    for i_input in range(model.n_var_rand):
        ind_base = i_input * n_density * n_sweeps
        for i_sweep in range(n_sweeps):
            ind_start = ind_base + i_sweep * n_density
            ind_end = ind_base + (i_sweep + 1) * n_density

            Q_all[ind_start:ind_end] = q_random[:, :, i_sweep]
            Q_all[ind_start:ind_end, i_input] = q_dense
            C_var[ind_start:ind_end] = [model.var_rand[i_input]] * n_density
            C_ind[ind_start:ind_end] = [i_sweep] * n_density

            ## Modify endpoints for infinite support
            if not isfinite(model.density.marginals[model.var_rand[i_input]].q(0)):
                Q_all[ind_start, i_input] = 1 / n_density / 10
            if not isfinite(model.density.marginals[model.var_rand[i_input]].q(1)):
                Q_all[ind_end - 1, i_input] = 1 - 1 / n_density / 10

    ## Assemble sampling plan
    df_pr = DataFrame(data=Q_all, columns=model.var_rand)
    df_rand = model.density.pr2sample(df_pr)
    df_rand[varname] = C_var
    df_rand[indname] = C_ind
    ## Construct outer-product DOE
    df_samp = model.var_outer(df_rand, df_det=df_det)

    if skip:
        ## Evaluation estimate
        runtime_est = model.runtime(df_samp.shape[0])
        if runtime_est > 0:
            print(
                "Estimated runtime for design with model ({0:1}):\n  {1:4.3} sec".format(
                    model.name, runtime_est
                )
            )
        else:
            print("Design runtime estimates unavailable; model has no timing data.")

        ## For autoplot
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            df_samp._plot_info = {"type": "sinew_inputs", "var": model.var_rand}

        ## Pass-through
        return df_samp

    ## Apply
    df_res = eval_df(model, df=df_samp, append=append)
    ## For autoplot
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        df_res._plot_info = {
            "type": "sinew_outputs",
            "var": model.var_rand,
            "out": model.out,
        }

    return df_res


ev_sinews = add_pipe(eval_sinews)

## Hybrid points for Sobol' indices
# --------------------------------------------------
@curry
def eval_hybrid(
    model,
    n=1,
    plan="first",
    df_det=None,
    varname="hybrid_var",
    seed=None,
    append=True,
    skip=False,
):
    r"""Hybrid points for Sobol' indices

    Use the "hybrid point" design (Sobol', 1999) to support estimating Sobol'
    indices. Use gr.tran_sobol() to post-process the results and compute
    estimates.

    Args:
        model (gr.Model): Model to evaluate; must have CopulaIndependence
        n (numeric): Number of points along each sweep
        plan (str): Sobol' index to compute; plan={"first", "total"}
        seed (int): Random seed to use
        df_det (DataFrame): Deterministic levels for evaluation; use "nom"
            for nominal deterministic levels.
        varname (str): Column name to give for sweep variable; default="hybrid_var"
        append (bool): Append results to conservative inputs?
        skip (bool): Skip evaluation of the functions?

    Returns:
        DataFrame: Results of evaluation or unevaluated design

    References:
        I.M. Sobol', "Sensitivity Estimates for Nonlinear Mathematical Models"
        (1999) MMCE, Vol 1.

    Examples:

        >>> import grama as gr
        >>> md = gr.make_cantilever_beam()
        >>> df_first = md >> gr.ev_hybrid(df_det="nom", plan="first")
        >>> df_first >> gr.tf_sobol()
        >>>
        >>> df_total = md >> gr.ev_hybrid(df_det="nom", plan="total")
        >>> df_total >> gr.tf_sobol()

    """
    ## Check invariants
    if not isinstance(model.density.copula, CopulaIndependence):
        raise ValueError(
            "model must have CopulaIndependence structure;\n"
            + "Sobol' indices only defined for independent variables"
        )

    ## Set seed only if given
    if seed is not None:
        set_seed(seed)

    if not isinstance(n, Integral):
        print("eval_hybrid() is rounding n...")
        n = int(n)

    ## Draw hybrid points
    X = random((n, model.n_var_rand))
    Z = random((n, model.n_var_rand))

    ## Reserve space
    Q_all = zeros((n * (model.n_var_rand + 1), model.n_var_rand))
    Q_all[:n] = X  # Base samples
    C_var = ["_"] * (n * (model.n_var_rand + 1))

    ## Interleave samples
    for i_in in range(model.n_var_rand):
        i_start = (i_in + 1) * n
        i_end = (i_in + 2) * n

        if plan == "first":
            Q_all[i_start:i_end, :] = Z
            Q_all[i_start:i_end, i_in] = X[:, i_in]
        elif plan == "total":
            Q_all[i_start:i_end, :] = X
            Q_all[i_start:i_end, i_in] = Z[:, i_in]
        else:
            raise ValueError("plan must be `first` or `total`")

        C_var[i_start:i_end] = [model.var_rand[i_in]] * n

    ## Construct sampling plan
    df_pr = DataFrame(data=Q_all, columns=model.var_rand)
    ## Convert samples to desired marginals
    df_rand = model.density.pr2sample(df_pr)
    df_rand[varname] = C_var
    ## Construct outer-product DOE
    df_samp = model.var_outer(df_rand, df_det=df_det)

    if skip:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            df_samp._meta = dict(
                type="eval_hybrid",
                varname=varname,
                plan=plan,
                var_rand=model.var_rand,
                out=model.out,
            )

        return df_samp

    df_res = eval_df(model, df=df_samp, append=append)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        df_res._meta = dict(
            type="eval_hybrid",
            varname=varname,
            plan=plan,
            var_rand=model.var_rand,
            out=model.out,
        )

    return df_res


ev_hybrid = add_pipe(eval_hybrid)



__all__ = [
    "tran_arrange",
    "tf_arrange",
    "tran_rename",
    "tf_rename",
    "tran_separate",
    "tf_separate",
    "tran_unite",
    "tf_unite",
    "tran_gather",
    "tf_gather",
    "tran_spread",
    "tf_spread",
    "tran_explode",
    "tf_explode",
    "convert_type"
]

import re
from .base import dfdelegate, symbolic_evaluation, flatten
from .. import add_pipe
from numpy import arange, nan
from pandas import Series, concat, melt, to_numeric, to_datetime


# ------------------------------------------------------------------------------
# Sorting
# ------------------------------------------------------------------------------


@dfdelegate
def tran_arrange(df, *args, **kwargs):
    """Calls `pandas.DataFrame.sort_values` to sort a DataFrame according to
    criteria.

    See:
    http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.sort_values.html

    For a list of specific keyword arguments for sort_values (which will be
    the same in arrange).

    Args:
        *args: Symbolic, string, integer or lists of those types indicating
            columns to sort the DataFrame by.

    Kwargs:
        **kwargs: Any keyword arguments will be passed through to the pandas
            `DataFrame.sort_values` function.
    """

    flat_args = [a for a in flatten(args)]

    series = [
        df[arg]
        if isinstance(arg, str)
        else df.iloc[:, arg]
        if isinstance(arg, int)
        else Series(arg)
        for arg in flat_args
    ]

    sorter = concat(series, axis=1).reset_index(drop=True)
    sorter = sorter.sort_values(sorter.columns.tolist(), **kwargs)
    return df.iloc[sorter.index, :].reset_index(drop=True)

tf_arrange = add_pipe(tran_arrange)


# ------------------------------------------------------------------------------
# Renaming
# ------------------------------------------------------------------------------


@dfdelegate
@symbolic_evaluation(eval_as_label=True)
def tran_rename(df, **kwargs):
    """Renames columns, where keyword argument values are the current names
    of columns and keys are the new names.

    Args:
        df (:obj:`pandas.DataFrame`): DataFrame passed in via `>>` pipe.

    Kwargs:
        **kwargs: key:value pairs where keys are new names for columns and
            values are current names of columns.
    """

    return df.rename(columns={v: k for k, v in kwargs.items()})

tf_rename = add_pipe(tran_rename)


# ------------------------------------------------------------------------------
# Elongate
# ------------------------------------------------------------------------------


@dfdelegate
@symbolic_evaluation(eval_as_label=["*"])
def tran_gather(df, key, values, *args, **kwargs):
    """
    Melts the specified columns in your DataFrame into two key:value columns.

    Args:
        key (str): Name of identifier column.
        values (str): Name of column that will contain values for the key.
        *args (str, int, symbolic): Columns to "melt" into the new key and
            value columns. If no args are specified, all columns are melted
            into they key and value columns.

    Kwargs:
        add_id (bool): Boolean value indicating whether to add a `"_ID"`
            column that will preserve information about the original rows
            (useful for being able to re-widen the data later).

    Example:
        diamonds >> gather('variable', 'value', ['price', 'depth','x','y','z']) >> head(5)

           carat      cut color clarity  table variable  value
        0   0.23    Ideal     E     SI2   55.0    price  326.0
        1   0.21  Premium     E     SI1   61.0    price  326.0
        2   0.23     Good     E     VS1   65.0    price  327.0
        3   0.29  Premium     I     VS2   58.0    price  334.0
        4   0.31     Good     J     SI2   58.0    price  335.0
    """

    if len(args) == 0:
        args = df.columns.tolist()
    else:
        args = [a for a in flatten(args)]

    if kwargs.get("add_id", False):
        df = df.assign(_ID=arange(df.shape[0]))

    columns = df.columns.tolist()
    id_vars = [col for col in columns if col not in args]
    return melt(df, id_vars, list(args), key, values)

tf_gather = add_pipe(tran_gather)


# ------------------------------------------------------------------------------
# Widen
# ------------------------------------------------------------------------------


def convert_type(df, columns):
    """
    Helper function that attempts to convert columns into their appropriate
    data type.
    """
    # taken in part from the dplython package
    out_df = df.copy()
    for col in columns:
        column_values = Series(out_df[col].unique())
        column_values = column_values[~column_values.isnull()]
        # empty
        if len(column_values) == 0:
            continue
        # boolean
        if set(column_values.values) < {"True", "False"}:
            out_df[col] = out_df[col].map({"True": True, "False": False})
            continue
        # numeric
        if to_numeric(column_values, errors="coerce").isnull().sum() == 0:
            out_df[col] = to_numeric(out_df[col], errors="ignore")
            continue
        # datetime
        if to_datetime(column_values, errors="coerce").isnull().sum() == 0:
            out_df[col] = to_datetime(
                out_df[col], errors="ignore", infer_datetime_format=True
            )
            continue

    return out_df


@dfdelegate
@symbolic_evaluation(eval_as_label=["*"])
def tran_spread(df, key, values, convert=False, fill=None):
    """
    Transforms a "long" DataFrame into a "wide" format using a key and value
    column.

    If you have a mixed datatype column in your long-format DataFrame then the
    default behavior is for the spread columns to be of type `object`, or
    string. If you want to try to convert dtypes when spreading, you can set
    the convert keyword argument in spread to True.

    Args:
        key (str, int, or symbolic): Label for the key column.
        values (str, int, or symbolic): Label for the values column.

    Kwargs:
        convert (bool): Boolean indicating whether or not to try and convert
            the spread columns to more appropriate data types.


    Example:
        widened = elongated >> spread(X.variable, X.value)
        widened >> head(5)

            _ID carat clarity color        cut depth price table     x     y     z
        0     0  0.23     SI2     E      Ideal  61.5   326    55  3.95  3.98  2.43
        1     1  0.21     SI1     E    Premium  59.8   326    61  3.89  3.84  2.31
        2    10   0.3     SI1     J       Good    64   339    55  4.25  4.28  2.73
        3   100  0.75     SI1     D  Very Good  63.2  2760    56   5.8  5.75  3.65
        4  1000  0.75     SI1     D      Ideal  62.3  2898    55  5.83   5.8  3.62
    """

    # Taken mostly from dplython package
    columns = df.columns.tolist()
    id_cols = [col for col in columns if not col in [key, values]]

    temp_index = ["" for i in range(len(df))]
    for id_col in id_cols:
        temp_index += df[id_col].map(str)

    out_df = df.assign(temp_index=temp_index)
    out_df = out_df.set_index("temp_index")
    spread_data = out_df[[key, values]]

    if not all(
        spread_data.groupby([spread_data.index, key]).agg("count").reset_index()[values]
        < 2
    ):
        raise ValueError("Duplicate identifiers")

    spread_data = spread_data.pivot(columns=key, values=values)

    if convert and (out_df[values].dtype.kind in "OSaU"):
        columns_to_convert = [col for col in spread_data if col not in columns]
        spread_data = convert_type(spread_data, columns_to_convert)

    if not (fill is None):
        spread_data.fillna(value=fill, inplace=True)

    out_df = out_df[id_cols].drop_duplicates()
    out_df = out_df.merge(spread_data, left_index=True, right_index=True).reset_index(
        drop=True
    )

    out_df = (out_df >> tf_arrange(id_cols)).reset_index(drop=True)

    return out_df

tf_spread = add_pipe(tran_spread)


# ------------------------------------------------------------------------------
# Separate columns
# ------------------------------------------------------------------------------


@dfdelegate
@symbolic_evaluation(eval_as_label=["*"])
def tran_separate(
    df,
    column,
    into,
    sep="[\W_]+",
    remove=True,
    convert=False,
    extra="drop",
    fill="right",
):
    """
    Splits columns into multiple columns.

    Args:
        df (pandas.DataFrame): DataFrame passed in through the pipe.
        column (str, symbolic): Label of column to split.
        into (list): List of string names for new columns.

    Kwargs:
        sep (str or list): If a string, the regex string used to split the
            column. If a list, a list of integer positions to split strings
            on.
        remove (bool): Boolean indicating whether to remove the original column.
        convert (bool): Boolean indicating whether the new columns should be
            converted to the appropriate type.
        extra (str): either `'drop'`, where split pieces beyond the specified
            new columns are dropped, or `'merge'`, where the final split piece
            contains the remainder of the original column.
        fill (str): either `'right'`, where `np.nan` values are filled in the
            right-most columns for missing pieces, or `'left'` where `np.nan`
            values are filled in the left-most columns.
    """

    assert isinstance(into, (tuple, list))

    if isinstance(sep, (tuple, list)):
        inds = [0] + list(sep)
        if len(inds) > len(into):
            if extra == "drop":
                inds = inds[: len(into) + 1]
            elif extra == "merge":
                inds = inds[: len(into)] + [None]
        else:
            inds = inds + [None]

        splits = df[column].map(
            lambda x: [
                str(x)[slice(inds[i], inds[i + 1])] if i < len(inds) - 1 else nan
                for i in range(len(into))
            ]
        )

    else:
        maxsplit = len(into) - 1 if extra == "merge" else 0
        splits = df[column].map(lambda x: re.split(sep, x, maxsplit))

    right_filler = lambda x: x + [nan for i in range(len(into) - len(x))]
    left_filler = lambda x: [nan for i in range(len(into) - len(x))] + x

    if fill == "right":
        splits = [right_filler(x) for x in splits]
    elif fill == "left":
        splits = [left_filler(x) for x in splits]

    for i, split_col in enumerate(into):
        df[split_col] = [x[i] if not x[i] == "" else nan for x in splits]

    if convert:
        df = convert_type(df, into)

    if remove:
        df.drop(column, axis=1, inplace=True)

    return df

tf_separate = add_pipe(tran_separate)


# ------------------------------------------------------------------------------
# Unite columns
# ------------------------------------------------------------------------------


@dfdelegate
@symbolic_evaluation(eval_as_label=["*"])
def tran_unite(df, colname, *args, **kwargs):
    """
    Does the inverse of `separate`, joining columns together by a specified
    separator.

    Any columns that are not strings will be converted to strings.

    Args:
        df (pandas.DataFrame): DataFrame passed in through the pipe.
        colname (str): the name of the new joined column.
        *args: list of columns to be joined, which can be strings, symbolic, or
            integer positions.

    Kwargs:
        sep (str): the string separator to join the columns with.
        remove (bool): Boolean indicating whether or not to remove the
            original columns.
        na_action (str): can be one of `'maintain'` (the default),
            '`ignore'`, or `'as_string'`. The default will make the new column
            row a `NaN` value if any of the original column cells at that
            row contained `NaN`. '`ignore'` will treat any `NaN` value as an
            empty string during joining. `'as_string'` will convert any `NaN`
            value to the string `'nan'` prior to joining.
    """

    to_unite = list([a for a in flatten(args)])
    sep = kwargs.get("sep", "_")
    remove = kwargs.get("remove", True)
    # possible na_action values
    # ignore: empty string
    # maintain: keep as np.nan (default)
    # as_string: becomes string 'nan'
    na_action = kwargs.get("na_action", "maintain")

    # print(to_unite, sep, remove, na_action)

    if na_action == "maintain":
        df[colname] = df[to_unite].apply(
            lambda x: nan if any(x.isnull()) else sep.join(x.map(str)), axis=1
        )
    elif na_action == "ignore":
        df[colname] = df[to_unite].apply(
            lambda x: sep.join(x[~x.isnull()].map(str)), axis=1
        )
    elif na_action == "as_string":
        df[colname] = df[to_unite].astype(str).apply(lambda x: sep.join(x), axis=1)

    if remove:
        df.drop(to_unite, axis=1, inplace=True)

    return df

tf_unite = add_pipe(tran_unite)


# ------------------------------------------------------------------------------
# Nesting
# ------------------------------------------------------------------------------


@dfdelegate
@symbolic_evaluation(eval_as_label=["*"])
def tran_explode(df, col, convert=False):
    """Lengthen DataFrame by exploding iterable entries in a column.

    If you have a mixed datatype column in your long-format DataFrame then the
    default behavior is for the spread columns to be of type `object`, or
    string. If you want to try to convert dtypes when spreading, you can set
    the convert keyword argument in spread to True.

    Note: Implemented in terms of the Pandas pd.DataFrame.explode() function.

    Args:
        col (str, int, or symbolic): Label for the column to explode.

    Kwargs:
        convert (bool): Boolean indicating whether or not to try and convert
            the spread columns to more appropriate data types.

    Returns:
        DataFrame:

    Example:

    """

    df_res = df.explode(col).reset_index(drop=True)

    if convert and (df_res[col].dtype.kind in "OSaU"):
        return convert_type(df_res, [col])
    return df_res

tf_explode = add_pipe(tran_explode)



# Import core
# --------------------------------------------------
from .tools import *
from .core import *

# Import dfply
# --------------------------------------------------
from .dfply import *

# Marginals uses make_symbolic decorator
from .marginals import *

# Add functionality to dfply
from .string_helpers import *
from .mutate_helpers import *

## Load plotnine
# --------------------------------------------------
from plotnine import *

## Load grama tools
# --------------------------------------------------
from .eval_defaults import *
from .tran_tools import *

from .comp_building import *
from .comp_metamodels import *
from .eval_random import *
from .eval_tail import *
from .eval_opt import *
from .eval_contour import *
from .tran_pivot import *
from .tran_shapley import *
from .tran_summaries import *
from .support import *
from .tran_is import *

from .fit_synonyms import *
from .plot_auto import *

## Load extras
# --------------------------------------------------
from .eval import *
from .fit import *
from .psdr import *
from .tran import *



__all__ = [
    "tran_count",
    "tf_count",
]

from .base import dfdelegate
from .. import add_pipe
from pandas import DataFrame

@dfdelegate
def tran_count(df, *args):
    r"""Count occurrences according to groups

    Count the number of rows within a specified grouping.

    Arguments
        df (DataFrame): Data to compute counts over.
        *args (str or gr.Intention()): Columns for grouping.

    Returns:
        DataFrame: Result of group counting.

    Examples:
        >>> import grama as gr
        >>> from grama.data import df_diamonds
        >>> DF = gr.Intention()
        >>> (
        >>>     # Single group variable
        >>>     df_diamonds
        >>>     >> gr.tf_count(DF.cut)
        >>> )
        >>>
        >>> (
        >>>     # Multiple group variables
        >>>     df_diamonds
        >>>     >> gr.tf_count(DF.cut, DF.clarity)
        >>> )

    """
    # Compute the count
    df_res = DataFrame(
        df.groupby([*args]).size()
    ).reset_index()
    # Change column name
    col = list(df_res.columns)
    col[-1] = "n"
    df_res.columns = col

    return df_res

tf_count = add_pipe(tran_count)



__all__ = [
    "tran_inner_join",
    "tf_inner_join",
    "tran_full_join",
    "tf_full_join",
    "tran_outer_join",
    "tf_outer_join",
    "tran_left_join",
    "tf_left_join",
    "tran_right_join",
    "tf_right_join",
    "tran_semi_join",
    "tf_semi_join",
    "tran_anti_join",
    "tf_anti_join",
    "tran_bind_rows",
    "tf_bind_rows",
    "tran_bind_cols",
    "tf_bind_cols",
]

from .base import dfdelegate
from .. import add_pipe
from pandas import concat


# ------------------------------------------------------------------------------
# SQL-style joins
# ------------------------------------------------------------------------------


def get_join_parameters(join_kwargs):
    """+
    Convenience function to determine the columns to join the right and
    left DataFrames on, as well as any suffixes for the columns.
    """

    by = join_kwargs.get("by", None)
    suffixes = join_kwargs.get("suffixes", ("_x", "_y"))
    if isinstance(by, tuple):
        left_on, right_on = by
    elif isinstance(by, list):
        by = [x if isinstance(x, tuple) else (x, x) for x in by]
        left_on, right_on = (list(x) for x in zip(*by))
    else:
        left_on, right_on = by, by
    return left_on, right_on, suffixes


@dfdelegate
def tran_inner_join(df, other, **kwargs):
    """
    Joins on values present in both DataFrames.

    Args:
        df (pandas.DataFrame): Left DataFrame (passed in via pipe)
        other (pandas.DataFrame): Right DataFrame

    Kwargs:
        by (str or list): Columns to join on. If a single string, will join
            on that column. If a list of lists which contain strings or
            integers, the right/left columns to join on.
        suffixes (list): String suffixes to append to column names in left
            and right DataFrames.

    Example:
        a >> inner_join(b, by='x1')

          x1  x2     x3
        0  A   1   True
        1  B   2  False
    """

    left_on, right_on, suffixes = get_join_parameters(kwargs)
    joined = df.merge(
        other, how="inner", left_on=left_on, right_on=right_on, suffixes=suffixes
    )
    return joined

tf_inner_join = add_pipe(tran_inner_join)


@dfdelegate
def tran_full_join(df, other, **kwargs):
    """
    Joins on values present in either DataFrame. (Alternate to `outer_join`)

    Args:
        df (pandas.DataFrame): Left DataFrame (passed in via pipe)
        other (pandas.DataFrame): Right DataFrame

    Kwargs:
        by (str or list): Columns to join on. If a single string, will join
            on that column. If a list of lists which contain strings or
            integers, the right/left columns to join on.
        suffixes (list): String suffixes to append to column names in left
            and right DataFrames.

    Example:
        a >> outer_join(b, by='x1')

          x1   x2     x3
        0  A  1.0   True
        1  B  2.0  False
        2  C  3.0    NaN
        3  D  NaN   True
    """

    left_on, right_on, suffixes = get_join_parameters(kwargs)
    joined = df.merge(
        other, how="outer", left_on=left_on, right_on=right_on, suffixes=suffixes
    )
    return joined

tf_full_join = add_pipe(tran_full_join)


@dfdelegate
def tran_outer_join(df, other, **kwargs):
    """
    Joins on values present in either DataFrame. (Alternate to `full_join`)

    Args:
        df (pandas.DataFrame): Left DataFrame (passed in via pipe)
        other (pandas.DataFrame): Right DataFrame

    Kwargs:
        by (str or list): Columns to join on. If a single string, will join
            on that column. If a list of lists which contain strings or
            integers, the right/left columns to join on.
        suffixes (list): String suffixes to append to column names in left
            and right DataFrames.

    Example:
        a >> full_join(b, by='x1')

          x1   x2     x3
        0  A  1.0   True
        1  B  2.0  False
        2  C  3.0    NaN
        3  D  NaN   True
    """

    left_on, right_on, suffixes = get_join_parameters(kwargs)
    joined = df.merge(
        other, how="outer", left_on=left_on, right_on=right_on, suffixes=suffixes
    )
    return joined

tf_outer_join = add_pipe(tran_outer_join)


@dfdelegate
def tran_left_join(df, other, **kwargs):
    """
    Joins on values present in in the left DataFrame.

    Args:
        df (pandas.DataFrame): Left DataFrame (passed in via pipe)
        other (pandas.DataFrame): Right DataFrame

    Kwargs:
        by (str or list): Columns to join on. If a single string, will join
            on that column. If a list of lists which contain strings or
            integers, the right/left columns to join on.
        suffixes (list): String suffixes to append to column names in left
            and right DataFrames.

    Example:
        a >> left_join(b, by='x1')

          x1  x2     x3
        0  A   1   True
        1  B   2  False
        2  C   3    NaN
    """

    left_on, right_on, suffixes = get_join_parameters(kwargs)
    joined = df.merge(
        other, how="left", left_on=left_on, right_on=right_on, suffixes=suffixes
    )
    return joined

tf_left_join = add_pipe(tran_left_join)


@dfdelegate
def tran_right_join(df, other, **kwargs):
    """
    Joins on values present in in the right DataFrame.

    Args:
        df (pandas.DataFrame): Left DataFrame (passed in via pipe)
        other (pandas.DataFrame): Right DataFrame

    Kwargs:
        by (str or list): Columns to join on. If a single string, will join
            on that column. If a list of lists which contain strings or
            integers, the right/left columns to join on.
        suffixes (list): String suffixes to append to column names in left
            and right DataFrames.

    Example:
        a >> right_join(b, by='x1')

          x1   x2     x3
        0  A  1.0   True
        1  B  2.0  False
        2  D  NaN   True
    """

    left_on, right_on, suffixes = get_join_parameters(kwargs)
    joined = df.merge(
        other, how="right", left_on=left_on, right_on=right_on, suffixes=suffixes
    )
    return joined

tf_right_join = add_pipe(tran_right_join)


@dfdelegate
def tran_semi_join(df, other, **kwargs):
    """
    Returns all of the rows in the left DataFrame that have a match
    in the right DataFrame.

    Args:
        df (pandas.DataFrame): Left DataFrame (passed in via pipe)
        other (pandas.DataFrame): Right DataFrame

    Kwargs:
        by (str or list): Columns to join on. If a single string, will join
            on that column. If a list of lists which contain strings or
            integers, the right/left columns to join on.

    Example:
        a >> semi_join(b, by='x1')

          x1  x2
        0  A   1
        1  B   2
    """

    left_on, right_on, suffixes = get_join_parameters(kwargs)
    if not right_on:
        right_on = [
            col_name
            for col_name in df.columns.values.tolist()
            if col_name in other.columns.values.tolist()
        ]
        left_on = right_on
    elif not isinstance(right_on, (list, tuple)):
        right_on = [right_on]
    other_reduced = other[right_on].drop_duplicates()
    joined = df.merge(
        other_reduced,
        how="inner",
        left_on=left_on,
        right_on=right_on,
        suffixes=("", "_y"),
        indicator=True,
    ).query('_merge=="both"')[df.columns.values.tolist()]
    return joined

tf_semi_join = add_pipe(tran_semi_join)


@dfdelegate
def tran_anti_join(df, other, **kwargs):
    """
    Returns all of the rows in the left DataFrame that do not have a
    match in the right DataFrame.

    Args:
        df (pandas.DataFrame): Left DataFrame (passed in via pipe)
        other (pandas.DataFrame): Right DataFrame

    Kwargs:
        by (str or list): Columns to join on. If a single string, will join
            on that column. If a list of lists which contain strings or
            integers, the right/left columns to join on.

    Example:
        a >> anti_join(b, by='x1')

          x1  x2
        2  C   3
    """

    left_on, right_on, suffixes = get_join_parameters(kwargs)
    if not right_on:
        right_on = [
            col_name
            for col_name in df.columns.values.tolist()
            if col_name in other.columns.values.tolist()
        ]
        left_on = right_on
    elif not isinstance(right_on, (list, tuple)):
        right_on = [right_on]
    other_reduced = other[right_on].drop_duplicates()
    joined = df.merge(
        other_reduced,
        how="left",
        left_on=left_on,
        right_on=right_on,
        suffixes=("", "_y"),
        indicator=True,
    ).query('_merge=="left_only"')[df.columns.values.tolist()]
    return joined

tf_anti_join = add_pipe(tran_anti_join)


# ------------------------------------------------------------------------------
# Binding
# ------------------------------------------------------------------------------


@dfdelegate
def tran_bind_rows(df, other, join="outer", ignore_index=False, reset=True):
    """
    Binds DataFrames "vertically", stacking them together. This is equivalent
    to `pd.concat` with `axis=0`.

    Args:
        df (pandas.DataFrame): Top DataFrame (passed in via pipe).
        other (pandas.DataFrame): Bottom DataFrame.

    Kwargs:
        join (str): One of `"outer"` or `"inner"`. Outer join will preserve
            columns not present in both DataFrames, whereas inner joining will
            drop them.
        ignore_index (bool): Indicates whether to consider pandas indices as
            part of the concatenation (defaults to `False`).
        reset (bool): Indicates whether to reset the dataframe index after
            bindin (defaults to `True`).

    """

    df = concat(
        [df.reset_index(drop=True), other.reset_index(drop=True)],
        join=join, ignore_index=ignore_index, axis=0, sort=False
    )

    if reset:
        return df.reset_index(drop=True)
    return df

tf_bind_rows = add_pipe(tran_bind_rows)


@dfdelegate
def tran_bind_cols(df, other, join="outer", ignore_index=False):
    """
    Binds DataFrames "horizontally". This is equivalent to `pd.concat` with
    `axis=1`.

    Args:
        df (pandas.DataFrame): Left DataFrame (passed in via pipe).
        other (pandas.DataFrame): Right DataFrame.

    Kwargs:
        join (str): One of `"outer"` or `"inner"`. Outer join will preserve
            rows not present in both DataFrames, whereas inner joining will
            drop them.
        ignore_index (bool): Indicates whether to consider pandas indices as
            part of the concatenation (defaults to `False`).
    """

    df = concat(
        [df.reset_index(drop=True), other.reset_index(drop=True)],
        join=join, ignore_index=ignore_index, axis=1, sort=False
    )
    return df

tf_bind_cols = add_pipe(tran_bind_cols)



__all__ = [
    "tran_shapley_cohort",
    "tf_shapley_cohort",
]

from grama import add_pipe, pipe
from itertools import chain, combinations
from numpy import all, number, sum, zeros, empty, NaN
from pandas import concat, DataFrame
from scipy.special import comb
from toolz import curry


## Helper
def powerset(iterable):
    s = list(iterable)

    return list(chain.from_iterable(combinations(s, r) for r in range(len(s) + 1)))


## Cohort Shapley
@curry
def tran_shapley_cohort(df, var=None, out=None, bins=20, inds=None):
    """Compute cohort shapley values

    Assess the impact of each variable on selected observations via cohort
    shapley [1]. Shapley values are a game-theoretic way to assess the
    importance of input variables (var) on each of a set of outputs (out). Since
    values are computed on each observation, cohort shapley can distinguish
    cases where a variable has a positive impact on one observation, and a
    negative impact on a different observation.

    Note that cohort shapley is combinatorialy expensive in the number of
    variables, and this expense is multiplied by the number of observations. Use
    with caution in cases of high dimensionality. Consider using the `inds`
    argument to analyze a small subset of your observations.

    Args:
        df (DataFrame): Variable and output data to analyze
        var (list of strings): Input variables
        out (list of strings): Outputs variables
        bins (integer): Number of "bins" to define coordinate refinement distance
        inds (iterable of indices or None): Indices of rows to analyze

    References:
        - [1] Mase, Owen, and Seiler, "Explaining black box decisions by Shapley cohort refinement" (2019) Arxiv

    Examples:
        >>> import grama as gr
        >>> from grama.data import df_stang
        >>> X = gr.Intention()
        >>> # Analyze all observations
        >>> (
        >>>     gr.tran_shapley_cohort(
        >>>         df_stang,
        >>>         var=["thick", "ang"],
        >>>         out=["E"],
        >>>     )
        >>>     >> gr.tf_bind_cols(df_stang)
        >>>     >> gr.tf_filter(X.E_thick < 0)
        >>> )
        >>> # Compute subset of values
        >>> (
        >>>     gr.tran_shapley_cohort(
        >>>         df_stang,
        >>>         var=["thick", "ang"],
        >>>         out=["E"],
        >>>         inds=(
        >>>             df_stang
        >>>             >> gr.tf_filter(X.thick > 0.08)
        >>>         ).index
        >>>     )
        >>>     >> gr.tf_bind_cols(df_stang)
        >>> )

    """
    ## Check invariants
    if not set(var).issubset(set(df.columns)):
        raise ValueError("var must be subset of df.columns")
    if not set(out).issubset(set(df.columns)):
        raise ValueError("out must be subset of df.columns")
    if len(set(var).intersection(set(out))) != 0:
        raise ValueError("var and out must have empty intersection")
    if inds is None:
        inds = range(df.shape[0])

    ## Setup
    s = df.shape[0]  # Number of observations (subjects)
    n = len(var)

    # Determine numeric and categorical columns
    var_numeric = list(df[var].select_dtypes(include=[number]).columns)
    var_cat = list(df[var].drop(columns=var_numeric).columns)

    # Compute distances for coordinate similarity
    df_dist = DataFrame(
        data={col: [(df[col].max() - df[col].min()) / bins] for col in var_numeric}
    )

    # Compute coordinate similarity boolean DataFrame
    df_sim = DataFrame(columns=["_i0", "_i1"] + list(df[var].columns))

    for i in range(s):
        ## Numeric comparison
        df_tmp = (
            df[var_numeric].iloc[i] - df[var_numeric].iloc[(i + 1) :]
        ).abs() <= df_dist[var_numeric].values
        ## Categorical comparison
        df_tmp[var_cat] = df[var_cat].iloc[i] == df[var_cat].iloc[(i + 1) :]
        ## Add subject indices
        df_tmp["_i0"] = [i] * df_tmp.shape[0]
        df_tmp["_i1"] = range((i + 1), s)

        ## Concatenate
        df_sim = concat((df_sim, df_tmp), axis=0, sort=False)

    # Internal functions
    def cohort_indices(t, varset):
        """Build set of cohort indices

        Args:
        t (integer): Target sample index
        varset (iterable): Variables for cohort refinement

        """
        if len(varset) == 0:
            return list(range(s))

        # Find all pairs similar along given variables
        flags_cohort = all(
            df_sim.drop(columns=["_i0", "_i1"])[[var[i] for i in varset]], axis=1
        ).values

        # Filter to pairs including target t
        df_tmp = df_sim[flags_cohort]
        df_cohort = df_tmp[(df_tmp["_i0"] == t) | (df_tmp["_i1"] == t)]

        # Consolidate index set
        return list(
            set(df_cohort["_i0"]).union(set(df_cohort["_i1"])).union(set((t,)))
        )

    def cohort_mean(t, varset):
        c = cohort_indices(t, varset)
        return df[out].iloc[c].mean().to_frame().T

    def cohort_shapley(j):
        """Cohort shapley for all observations, single variable
        """
        poset = powerset(set(range(n)).difference({j}))
        data = zeros((s, len(out)))
        df_tmp = DataFrame(columns=out, data=data)

        for p in poset:
            den = n * comb(n - 1, len(p))

            for t in range(s):
                if t in inds:
                    t1 = cohort_mean(t, list(set(p).union({j})))
                    t0 = cohort_mean(t, p)

                    df_tmp.iloc[t] = df_tmp.iloc[t] + (t1 - t0).loc[0] / den
                else:
                    df_tmp.iloc[t] = NaN

        return df_tmp

    ## Compute cohort shapley over all variables
    df_res = DataFrame()
    for j in range(n):
        df_tmp = cohort_shapley(j)
        df_tmp.columns = [df_tmp.columns[i] + "_" + var[j] for i in range(len(out))]

        df_res = concat((df_res, df_tmp), axis=1)

    return df_res


tf_shapley_cohort = add_pipe(tran_shapley_cohort)



__all__ = [
    "tran_head",
    "tf_head",
    "tran_tail",
    "tf_tail",
    "tran_sample",
    "tf_sample",
    "tran_distinct",
    "tf_distinct",
    "tran_row_slice",
    "tf_row_slice",
    "tran_filter",
    "tf_filter",
    "tran_top_n",
    "tf_top_n",
    "tran_pull",
    "tf_pull",
    "tran_dropna",
    "tf_dropna",
]

import warnings
from .base import dfdelegate, symbolic_evaluation, group_delegation
from .. import add_pipe
from numpy import array, ones
from pandas import Series


# ------------------------------------------------------------------------------
# `head` and `tail`
# ------------------------------------------------------------------------------


@dfdelegate
def tran_head(df, n=5):
    return df.head(n)

tf_head = add_pipe(tran_head)


@dfdelegate
def tran_tail(df, n=5):
    return df.tail(n)

tf_tail = add_pipe(tran_tail)


# ------------------------------------------------------------------------------
# Sampling
# ------------------------------------------------------------------------------


@dfdelegate
def tran_sample(df, *args, **kwargs):
    return df.sample(*args, **kwargs).reset_index(drop=True)

tf_sample = add_pipe(tran_sample)


@group_delegation
@symbolic_evaluation(eval_as_label=["*"])
def tran_distinct(df, *args, **kwargs):
    if not args:
        return df.drop_duplicates(**kwargs).reset_index(drop=True)
    return df.drop_duplicates(list(args), **kwargs).reset_index(drop=True)

tf_distinct = add_pipe(tran_distinct)


@dfdelegate
def tran_row_slice(df, indices):
    if isinstance(indices, (tuple, list)):
        indices = array(indices)
    if isinstance(indices, int):
        indices = array([indices])
    if isinstance(indices, Series):
        indices = indices.values

    if indices.dtype == bool:
        return df.loc[indices, :].reset_index(drop=True)
    return df.iloc[indices, :].reset_index(drop=True)

tf_row_slice = add_pipe(tran_row_slice)


# ------------------------------------------------------------------------------
# Filtering/masking
# ------------------------------------------------------------------------------


@dfdelegate
def mask(df, *args):
    mask = Series(ones(df.shape[0], dtype=bool))
    for arg in args:
        if arg.dtype != bool:
            raise Exception("Arguments must be boolean.")
        mask = mask & arg.reset_index(drop=True)
    return df[mask.values].reset_index(drop=True)

tran_filter = mask  # alias for mask()

tf_filter = add_pipe(tran_filter)


@dfdelegate
def tran_top_n(df, n=None, ascending=True, col=None):
    if not n:
        raise ValueError("n must be specified")
    if not isinstance(col, Series):
        col = df.columns[-1]
    else:
        col = col._name
    index = df[[col]].copy()
    index["ranks"] = index[col].rank(ascending=ascending)
    index = index[index["ranks"] >= index["ranks"].nlargest(n).min()]
    return df.reindex(index.index)

tf_top_n = add_pipe(tran_top_n)


@dfdelegate
def tran_pull(df, col=-1):
    if not isinstance(col, Series):
        col = df.columns[-1]
    else:
        col = col._name

    return df[col]

tf_pull = add_pipe(tran_pull)


@dfdelegate
def tran_dropna(df, how="any", subset=None):
   return df.dropna(how=how, subset=subset).reset_index(drop=True)

tf_dropna = add_pipe(tran_dropna)



## grama core functions
# Zachary del Rosario, March 2019

__all__ = [
    "CopulaIndependence",
    "CopulaGaussian",
    "Domain",
    "Density",
    "Function",
    "FunctionModel",
    "FunctionVectorized",
    "Model",
    "NaN",
]

import copy
import networkx as nx
import warnings
from grama import pipe, valid_dist, param_dist
from .tools import tran_outer
from abc import ABC, abstractmethod
from itertools import chain
from numpy import ones, zeros, triu_indices, eye, array, Inf, NaN, sqrt, \
    dot, diag, isfinite, prod, exp
from numpy import min as npmin
from numpy import max as npmax
from numpy.linalg import cholesky, det, inv
from numpy.random import random, multivariate_normal
from numpy.random import seed as set_seed
from pandas import DataFrame, concat
from scipy.linalg import det, LinAlgError, solve
from scipy.optimize import root_scalar
from scipy.stats import norm, gaussian_kde
from toolz import curry


## Package settings
RUNTIME_LOWER = 1  # Cutoff threshold for runtime messages

## Core functions
##################################################
# Function class
class Function:
    """Parent class for functions.

    A function specifies its own inputs and outputs; these are subsets of the
    full model's inputs and outputs.

    """

    def __init__(self, func, var, out, name, runtime):
        """Function constructor

        Construct a grama function. Generally not called directly; preferred
        usage is through gr.comp_function().

        Args:
            func (function): Function mapping X^d -> X^r
            var (list(str)): Named variables; must match order of X^d
            out (list(str)): Named outputs; must match order of X^r
            name (str): Function name
            runtime (numeric): Estimated single-eval runtime (in seconds)

        Returns:
            gr.Function: grama function

        """
        self.func = func
        self.var = var
        self.out = out
        self.name = name
        self.runtime = runtime

    def copy(self):
        """Make a copy"""
        func_new = Function(
            copy.deepcopy(self.func),
            copy.deepcopy(self.var),
            copy.deepcopy(self.out),
            copy.deepcopy(self.name),
            runtime=self.runtime,
        )
        return func_new

    def eval(self, df):
        """Evaluate function

        Evaluate a grama function; loops over dataframe rows. Intended for
        internal use.

        Args:
        df (DataFrame): Input values to evaluate

        Returns:
            DataFrame: Result values

        """
        ## Check invariant; model inputs must be subset of df columns
        if not set(self.var).issubset(set(df.columns)):
            raise ValueError(
                "Model function `{}` var not a subset of given columns".format(
                    self.name
                )
            )

        ## Set up output
        n_rows = df.shape[0]
        results = zeros((n_rows, len(self.out)))
        for ind in range(n_rows):
            results[ind] = self.func(df.loc[ind, self.var])

        ## Package output as DataFrame
        return DataFrame(data=results, columns=self.out)

    def summary(self):
        """Returns a summary string
        """
        return "{0:}: {1:} -> {2:}".format(self.name, self.var, self.out)


class FunctionVectorized(Function):
    def eval(self, df):
        """Evaluate function; DataFrame vectorized

        Evaluate grama function. Assumes function is vectorized over dataframes.

        Args:
            df (DataFrame): Input values to evaluate

        Returns:
            DataFrame: Result values

        """
        df_res = self.func(df)
        return df_res[self.out]

    def copy(self):
        """Make a copy"""
        func_new = FunctionVectorized(
            self.func, self.var, self.out, self.name, self.runtime
        )
        return func_new


class FunctionModel(Function):
    """gr.Model as gr.Function
    """

    def __init__(self, md, ev=None, var=None, out=None):
        """Model-Function constructor

        Construct a grama function from a model. Generally not called directly;
        preferred usage is through gr.comp_model().

        Args:
            md (gr.Model): Grama model
            ev (function): Evaluation function for model; must have signature
                ev(md, df); must take df with columns matching given `var`
            var (list): Variables used by ev() to evaluate md. Ignored if
                default ev used.
            out (list): Outputs returned by ev(). Ignored if default ev used.

        Returns:
            gr.Function: grama function

        """
        self.model = md

        ## Construct default evaluator
        if ev is None:

            def _ev(md, df):
                df_res = md.evaluate_df(df)
                return df_res[md.out]

            self.ev = _ev
            self.var = self.model.var
            self.out = self.model.out

        ## Use given evaluator
        else:
            self.ev = ev
            self.var = var
            self.out = out

        ## Copy model data
        self.runtime = md.runtime(1)
        self.name = copy.copy(md.name)

    def eval(self, df):
        """Evaluate function; DataFrame vectorized

        Evaluate grama model as a function. Modify the parameters before

        Args:
            df (DataFrame): Input values to evaluate

        Returns:
            DataFrame: Result values

        """
        return self.ev(self.model, df)

    def copy(self):
        """Make a copy"""
        func_new = FunctionModel(self.model, ev=self.ev, var=self.var, out=self.out)
        return func_new


# Domain parent class
class Domain:
    """Parent class for input domains

    The domain defines constraints on the variables. Together with a model's
    functions, it defines the mathematical domain of a model.

    """

    def __init__(self, bounds=None, feasible=None):
        """Constructor

        Construct a grama domain. Generally not called directly; preferred usage
        is through gr.comp_bounds().

        Args:
        bounds (dict): Variable bounds, given as {"var": [Lower, Upper]}
            dict entires
        feasible (function): Vectorized function mapping variables to bool
            NOT IMPLEMENTED

        Returns:
            gr.Domain: grama domain

        @pre isinstance(bounds, dict)
        @pre (feasible is function) | (feasible is None)
        @pre len(bounds) == n where model.function:R^n -> R^m

        """
        if bounds is None:
            bounds = {}

        self.bounds = bounds
        self.feasible = feasible
        self.update()

    def update(self):
        self.var = list(self.bounds.keys())

    def copy(self):
        new_domain = Domain(
            bounds=copy.deepcopy(self.bounds), feasible=copy.deepcopy(self.feasible)
        )

        return new_domain

    def get_bound(self, var):
        if var in self.bounds.keys():
            return (self.bounds[var][0], self.bounds[var][1])
        return (-Inf, +Inf)

    def get_width(self, var):
        if var in self.bounds.keys():
            return self.bounds[var][1] - self.bounds[var][0]
        return +Inf

    def get_nominal(self, var):
        if var in self.bounds.keys():
            if isfinite(self.bounds[var][0]):
                if isfinite(self.bounds[var][1]):
                    return 0.5 * (self.bounds[var][1] + self.bounds[var][0])
                return self.bounds[var][0]
            if isfinite(self.bounds[var][1]):
                return self.bounds[var][1]
            return NaN
        return NaN

    def bound_summary(self, var):
        if var in self.bounds.keys():
            return "{0:}: [{1:}, {2:}]".format(
                var, self.bounds[var][0], self.bounds[var][1],
            )
        return "{0:}: (unbounded)".format(var)


## Copula base class
class Copula(ABC):
    """Parent class for copulas
    """

    @abstractmethod
    def __init__(self):
        raise NotImplementedError

    @abstractmethod
    def copy(self):
        raise NotImplementedError

    @abstractmethod
    def sample(self, n=1):
        r"""Draw a sample of a given size
        """
        raise NotImplementedError

    @abstractmethod
    def l(self, u):
        r"""Copula density
        """
        raise NotImplementedError

    @abstractmethod
    def u2z(self, u):
        r"""Transform from [0, 1]^d to sample space
        """
        raise NotImplementedError

    @abstractmethod
    def z2u(self, z):
        r"""Transform from sample space to [0, 1]^d
        """
        raise NotImplementedError

    @abstractmethod
    def dudz(self, z):
        r"""Jacobian of copula transform
        """
        raise NotImplementedError

    @abstractmethod
    def summary(self):
        raise NotImplementedError


class CopulaIndependence(Copula):
    def __init__(self, var_rand):
        """Constructor

        Args:
            var_rand (ind): Number of random variables

        Returns:
            gr.CopulaIndependence: Independence copula

        """
        self.var_rand = var_rand

    def copy(self):
        """Copy

        Returns:
            gr.CopulaIndependence: Copy of present copula

        """
        cop = CopulaIndependence(var_rand=self.var_rand)

        return cop

    def sample(self, n=1, seed=None):
        """Draw samples from copula

        Args:
            n (int): Number of samples
            seed (int): Random seed

        Returns:
            DataFrame: Independent samples
        """
        ## Set seed only if given
        if seed is not None:
            set_seed(seed)

        return DataFrame(data=random((n, len(self.var_rand))), columns=self.var_rand)

    def l(self, u):
        """Density function

        Args:
            u (array-like):

        Returns:
            array: Copula density values

        """
        return ones(u.shape[0])

    def u2z(self, u):
        """Transform to standard-normal space

        Args:
            u (array-like):

        Returns:
            array:

        """
        return norm.ppf(u)

    def z2u(self, z):
        """Transform to uniform-marginal space

        Args:
            z (array-like):

        Returns:
            array:

        """
        return norm.cdf(z)

    def dudz(self, z):
        """Jacobian

        Args:
            z (array-like):

        Returns:
            array:

        """
        return diag(norm.pdf(z))

    def summary(self):
        return "Independence copula"


class CopulaGaussian(Copula):
    def __init__(self, var_rand, df_corr):
        """Constructor

        Args:
            self (gr.CopulaGaussian):
            df_corr (DataFrame): Pairwise correlations provided as
                off-diagonal upper-triangular entries. Must have columns
                ["var1", "var2", "corr"]. Zero-correlation entires must
                be provided.

        Returns:
            gr.CopulaGaussian: Gaussian copula

        """
        ## Sort correlations along given var_rand
        sorter_index = dict(zip(var_rand, range(len(var_rand))))
        df_corr["var1_rank"] = df_corr["var1"].map(sorter_index)
        df_corr["var2_rank"] = df_corr["var2"].map(sorter_index)
        df_corr.sort_values(
            ["var1_rank", "var2_rank"], ascending=[True, True], inplace=True
        )
        df_corr.drop(["var1_rank", "var2_rank"], axis=1, inplace=True)

        n_var_rand = len(var_rand)
        Ind_upper = triu_indices(n_var_rand, 1)

        ## Check invariants
        if len(Ind_upper[0]) != len(df_corr["corr"]):
            raise ValueError("Invalid set of correlations provided")

        ## Build correlation structure
        Sigma = eye(n_var_rand)
        Sigma[Ind_upper] = df_corr["corr"].values
        Sigma = Sigma + (Sigma - eye(n_var_rand)).T
        try:
            Sigma_h = cholesky(Sigma)
            Sigma_i = inv(Sigma)
        except LinAlgError:
            warnings.warn(
                "Correlation structure is not positive-definite; copula transforms not available",
                RuntimeWarning
            )
            Sigma_h = None
            Sigma_i = None

        ## Build density quantities

        self.df_corr = df_corr
        self.var_rand = var_rand
        self.Sigma = Sigma
        self.Sigma_h = Sigma_h
        self.Sigma_i = Sigma_i
        self.det = det(Sigma)

    def copy(self):
        """Copy

        Args:
            self (gr.CopulaGaussian):

        Returns:
            gr.CopulaGaussian:
        """
        cop = CopulaGaussian(self.var_rand, self.df_corr.copy())

        return cop

    def sample(self, n=1, seed=None):
        """Draw samples from copula

        Draw samples according to gaussian copula dependence structure.

        Args:
            self (gr.CopulaGaussian):
            n (int): Number of samples to draw

        Returns:
            array: Copula samples

        """
        ## Set seed only if given
        if seed is not None:
            set_seed(seed)

        ## Generate correlated samples
        gaussian_samples = multivariate_normal(
            mean=[0] * len(self.var_rand), cov=self.Sigma, size=n
        )
        ## Convert to uniform marginals
        quantiles = valid_dist["norm"].cdf(gaussian_samples)

        return DataFrame(data=quantiles, columns=self.var_rand)

    def l(self, u):
        """Copula density function

        Args:
            u (array-like):

        Returns:
            array:

        """
        n_obs = u.shape[0]
        n_dim = u.shape[1]
        l_values = zeros(n_obs)

        for i in range(n_obs):
            v = norm.ppf(u[i, :])
            l_values[i] = exp(
                -0.5*dot(v, dot( self.Sigma_i - eye(n_dim), v))
            ) / sqrt(self.det)

        return l_values

    def u2z(self, u):
        """Transform to standard-normal space

        Args:
            u (array-like):

        Returns:
            array:

        """
        N = norm.ppf(u)
        Z = solve(self.Sigma_h, N)

        return Z

    def z2u(self, z):
        """Transform to uniform-marginal space

        Args:
            z (array-like):

        Returns:
            array:

        """
        return norm.cdf(dot(self.Sigma_h, z))

    def dudz(self, z):
        """Jacobian

        Args:
            z (array-like):

        Returns:
            array:

        """
        return dot(self.Sigma_h.T, diag(norm.pdf(dot(self.Sigma_h, z))))

    def summary(self):
        return "Gaussian copula with correlations:\n{}".format(self.df_corr)


## Density parent class
class Density:
    """Parent class for joint densities

    The density is defined for all the random variables; therefore it explicitly
    defines the list of random variables, and together implicitly defines the
    deterministic variables via

        domain.var + [functions.var] - density.marginals.keys()

    """

    def __init__(self, marginals=None, copula=None):
        """Constructor

        Construct a grama density. Generally not called directly; preferred
        usage is through gr.comp_marginals() and gr.comp_copula().

        Args:
            marginals (dict): Dictionary of gr.Marginal objects
            copula (gr.Copula): Copula object

        Returns:
            gr.Density: grama density

        """
        self.marginals = marginals
        self.copula = copula

    def copy(self):
        try:
            new_marginals = {}
            for key, value in self.marginals.items():
                new_marginals[key] = self.marginals[key].copy()

        except AttributeError:
            new_marginals = {}

        try:
            new_copula = self.copula.copy()
        except AttributeError:
            new_copula = None

        new_density = Density(marginals=new_marginals, copula=new_copula)

        return new_density

    def l(self, df):
        r"""Evaluate PDF

        Evaluate the PDF of the density.

        Args:
            df (DataFrame): Values

        """
        # Get variable names
        var = [key for key, _ in self.marginals.items()]
        df_u = self.sample2pr(df)[var]
        # Evaluate copula density
        l_copula = self.copula.l(df_u.values)
        # Evaluate marginal densities
        L_marginals = zeros((df.shape[0], len(var)))
        for i, v in enumerate(var):
            L_marginals[:, i] = self.marginals[v].l(df[v])
        l_marginals = prod(L_marginals, axis=1)

        return l_copula * l_marginals


    def pr2sample(self, df_prval):
        """Convert CDF probabilities to samples

        Convert random variable CDF probabilities to random variable samples.
        Ignores dependence structure.

        Args:
            df_prval (DataFrame): Values \in [0,1]

        Returns:
            DataFrame: Variable samples (quantiles)

        @pre df_prval.shape[1] == len(self.var_rand)
        @post result.shape[1] == len(self.var_rand)

        """
        try:
            var_rand = list(self.marginals.keys())
        except AttributeError:
            var_rand = []

        ## Empty case
        if len(var_rand) == 0:
            return DataFrame()

        ## Variables to convert
        var_comp = list(set(var_rand).intersection(set(df_prval.columns)))
        if len(var_comp) == 0:
            raise ValueError(
                "Intersection of df_prval.columns and var_rand must be nonempty"
            )

        samples = zeros(df_prval[var_comp].shape)
        ## Ensure correct column ordering
        prval = df_prval[var_comp].values

        ## Apply appropriate marginal
        for ind in range(len(var_comp)):
            ## Map with inverse density
            var = var_comp[ind]
            samples[:, ind] = self.marginals[var].q(prval[:, ind])

        return DataFrame(data=samples, columns=var_comp)

    def sample2pr(self, df_sample):
        """Convert samples to CDF probabilities

        Convert random variable samples to CDF probabilities. Ignores dependence
        structure.

        Args:
            df_sample (DataFrame): Values \in [0,1]

        Returns:
            DataFrame: Variable samples (quantiles)

        @pre df_sample.shape[1] == len(self.var_rand)
        @post result.shape[1] == len(self.var_rand)

        """
        try:
            var_rand = list(self.marginals.keys())
        except AttributeError:
            var_rand = []

        ## Empty case
        if len(var_rand) == 0:
            return DataFrame()

        ## Variables to convert
        var_comp = list(set(var_rand).intersection(set(df_sample.columns)))
        if len(var_comp) == 0:
            raise ValueError(
                "Intersection of df_sample.columns and var_rand must be nonempty"
            )

        prval = zeros(df_sample[var_comp].shape)
        ## Ensure correct column ordering
        sample = df_sample[var_comp].values

        ## Apply appropriate marginal
        for ind in range(len(var_comp)):
            ## Map with inverse density
            var = var_comp[ind]
            prval[:, ind] = self.marginals[var].p(sample[:, ind])

        return DataFrame(data=prval, columns=var_comp)

    def sample(self, n=1, seed=None):
        """Draw samples from joint density

        Draw samples according to joint density using marginal and copula
        information.

        Args:
            n (int): Number of samples to draw
            seed (int): random seed to use

        Returns:
            DataFrame: Joint density samples

        """
        if not (self.copula is None):
            df_pr = self.copula.sample(n=n, seed=seed)
        else:
            raise ValueError(
                "\n"
                + "Present model copula must be defined for sampling.\n"
                + "Use CopulaIndependence only when inputs can be guaranteed\n"
                + "independent. See the Documentation chapter on Random\n"
                + "Variable Modeling for more information."
            )
        return self.pr2sample(df_pr)

    def summary_marginal(self, var):
        return "{0:}: {1:}".format(var, self.marginals[var].summary())

    def summary_copula(self):
        if not (self.copula is None):
            return self.copula.summary()
        return None


# Model parent class
class Model:
    """Parent class for grama models.
    """

    def __init__(
        self, name=None, functions=None, domain=None, density=None,
    ):
        r"""Constructor

        Construct a grama model. Generally called without arguments; suggested
        procedure is to use gr.building tools to build up the model.

        Args:
            name (string): Name of model
            functions (list(gr.function)): Define the model mapping
                f(x) : R^n_in -> R^n_out along with function input and output names
            domain (gr.Domain): Model domain
            density (gr.Density): Model density

        Returns:
            gr.Model: grama model

        @pre len(domain.var) == n_in
        @pre len(out) == n_out
        @pre isinstance(domain, domain_)
        @pre isinstance(density, density_) || (density is None)

        Examples:

            >>> import grama as gr
            >>> print(gr.valid_dist.keys()) # Supported distributions
            >>> md = gr.Model() >> \
            >>>     gr.cp_function(
            >>>         lambda x: x[0] + x[1],
            >>>         var=["x0", "x1"],
            >>>         out=1
            >>>     ) >> \
            >>>     gr.cp_marginals(
            >>>         x0={"dist": "uniform", "loc": 0, "scale": 1}
            >>>     ) >> \
            >>>     gr.cp_bounds(x1=(0, 1))

        """
        if functions is None:
            functions = []
        if domain is None:
            domain = Domain()
        if density is None:
            density = Density()

        self.name = name
        self.functions = functions
        self.domain = domain
        self.density = density

        self.update()

    def update(self):
        """Update model public attributes based on functions, domain, and density.

        The variables and parameters are implicitly defined by the model
        attributes. For internal use.

        - self.functions defines the full list of inputs
        - self.domain defines the constraints on the model's domain
        - self.density defines the random variables

        """
        ## Initialize
        self.domain.update()
        self.var = self.domain.var.copy()
        self.out = []

        ## Construct var and out, respecting DAG properties
        for fun in self.functions:
            self.var = list(set(self.var).union(set(fun.var).difference(set(self.out))))

            self.out = list(set(self.out).union(set(fun.out)))

        try:
            self.var_rand = list(self.density.marginals.keys())
        except AttributeError:
            self.var_rand = []
        self.var_det = list(set(self.var).difference(self.var_rand))

        ## TODO parameters

        ## Convenience constants
        self.n_var = len(self.var)
        self.n_var_rand = len(self.var_rand)
        self.n_var_det = len(self.var_det)
        self.n_out = len(self.out)

    def runtime(self, n):
        """Estimate runtime

        Estimate the total runtime to evaluate n observations.

        Args:
            self (gr.Model):
            n (int): Number of observations

        Returns:
            float: Estimated runtime, in seconds

        """
        rate = 0
        for fun in self.functions:
            rate = rate + fun.runtime

        return rate * n

    def runtime_message(self, df):
        """Runtime message

        Estimate total runtime based on proposed DataFrame, prepare a message
        for console print. Print no message if runtime is negligible.

        Args:
            self (gr.Model):
            df (DataFrame): Data to evaluate

        Returns:
            str: Runtime message

        """
        runtime = self.runtime(df.shape[0])

        if runtime < RUNTIME_LOWER:
            return None
        return "Estimated runtime: {0:3.4f} sec".format(runtime)

    def det_nom(self):
        """Return nominal conditions for deterministic variables

        Returns:
            DataFrame: Nominal values for deterministic variables

        """
        data = {}

        for var in self.var_det:
            data[var] = [self.domain.get_nominal(var)]

        return DataFrame(data=data)

    def evaluate_df(self, df):
        """Evaluate function using an input dataframe

        Args:
            df (DataFrame): Variable values at which to evaluate model functions

        Returns:
            DataFrame: Output results

        """
        ## Check invariant; model inputs must be subset of df columns
        var_diff = set(self.var).difference(set(df.columns))
        if len(var_diff) != 0:
            raise ValueError(
                "Model inputs not a subset of given columns;\n"
                + "missing var = {}".format(var_diff)
            )

        df_tmp = df.copy().drop(self.out, axis=1, errors="ignore")
        ## Evaluate each function
        for func in self.functions:
            ## Concatenate to make intermediate results available
            df_tmp = concat((df_tmp, func.eval(df_tmp)), axis=1)

        return df_tmp[self.out]

    def var_outer(self, df_rand, df_det=None):
        """Outer product of random and deterministic samples

        Args:
            df_rand (DataFrame) random variable samples
            df_det (DataFrame) deterministic variable samples
                set to "nom" for nominal evaluation

        Returns:
            DataFrame: Outer product of samples

        """
        ## Pass-through if no var_det
        if self.n_var_det == 0:
            return df_rand

        ## Error-throwing default value
        if df_det is None:
            raise ValueError("df_det must be DataFrame or 'nom'")
        ## String shortcut
        elif isinstance(df_det, str):
            if df_det == "nom":
                df_det = self.det_nom()
            else:
                raise ValueError("df_det shortcut string invalid")
        ## DataFrame
        else:
            ## Check invariant; model inputs must be subset of df columns
            if not set(self.var_det).issubset(set(df_det.columns)):
                raise ValueError("model.var_det not a subset of given columns")

        ## Pass-through if no var_rand
        if self.n_var_rand == 0:
            return df_det

        ## Outer product if both det and rand exist
        return tran_outer(df_rand, df_det)

    ## Sample transforms
    # --------------------------------------------------
    def x2z(self, x):
        r"""Transform to standard normal space

        Transform a single vector of random variable values to standard normal
        space.

        Args:
            x (array): Single vector of values in var_rand. Order of entries
                must match self.var_rand

        Returns:
            array: Single vector of values transformed to standard normal space

        """
        ## Transform to uniform
        u = zeros(self.n_var_rand)
        for i in range(self.n_var_rand):
            u[i] = self.density.marginals[self.var_rand[i]].p(x[i])
        ## Transform to standard normal
        z = self.density.copula.u2z(u)

        return z

    def z2x(self, z):
        r"""Transform to random variable space

        Transform a single vector of normal values to the model's random
        variable space.

        Args:
            z (array): Single vector of standard normal values. Order of entries
                must match self.var_rand

        Returns:
            array: Single vector of values transformed to model random variable space

        """
        ## Correlate and map to uniform
        u = self.density.copula.z2u(z)
        ## Transform per marginal
        x = zeros(self.n_var_rand)
        for i in range(self.n_var_rand):
            x[i] = self.density.marginals[self.var_rand[i]].q(u[i])

        return x

    def dxdz(self, z):
        r"""Inverse transform jacobian

        Compute jacobian of the inverse transform X = phi^{-1}(Z)

        Args:
            z (array): Single vector of standard normal values. Order of entries
                must match self.var_rand

        Returns:
            2d array: Jacobian of inverse transform

        """
        ## Setup
        dudz = self.density.copula.dudz(z)

        x = self.z2x(z)
        F = zeros(self.n_var_rand)
        for i in range(self.n_var_rand):
            F[i] = 1 / self.density.marginals[self.var_rand[i]].l(x[i])

        return dot(dudz, diag(F))

    ## Sample transforms; DataFrame
    # --------------------------------------------------
    def rand2norm(self, df):
        r"""Transform random samples to standard normal space

        Transform a DataFrame of random variable samples to standard normal space

        Args:
            df (DataFrame): Random variable samples; must have columns for all
                of self.var_rand.

        Returns:
            DataFrame: Samples in standard-normal space
        """
        ## Check invariants
        if not set(self.var_rand).issubset(set(df.columns)):
            raise ValueError("model.var_rand must be subset of df.columns")

        data = zeros((df.shape[0], self.n_var_rand))
        for i in range(df.shape[0]):
            data[i] = self.x2z(df[self.var_rand].iloc[i].values)

        return DataFrame(data=data, columns=self.var_rand)

    def norm2rand(self, df):
        r"""Transform standard normal samples to model random variable space

        Transform a DataFrame of standard normal samples to model random
        variable space.

        Args:
            df (DataFrame): Random variable samples; must have columns for all
                of self.var_rand.

        Returns:
            DataFrame: Samples in standard-normal space

        """
        ## Check invariants
        if not set(self.var_rand).issubset(set(df.columns)):
            raise ValueError("model.var_rand must be subset of df.columns")

        data = zeros((df.shape[0], self.n_var_rand))
        for i in range(df.shape[0]):
            data[i] = self.z2x(df[self.var_rand].iloc[i].values)

        return DataFrame(data=data, columns=self.var_rand)

    def name_corr(self):
        """Name the correlation elements
        """
        raise NotImplementedError
        ## Build matrix of names
        corr_mat = []
        for ind in range(self.n_in):
            corr_mat.append(
                list(map(lambda s: s + "," + self.domain.var[ind], self.domain.var))
            )

        ## Access matrix of names
        corr_names = dict()
        corr_ind = triu_indices(self.n_in, 1)
        for knd in range(len(corr_ind[0])):
            ind = corr_ind[0][knd]
            jnd = corr_ind[1][knd]
            corr_names["corr_" + str(knd)] = corr_mat[ind][jnd]

        return corr_names

    ## Infrastructure
    # -------------------------
    def copy(self):
        """Make a copy of this model
        """
        new_model = Model(
            name=self.name,
            functions=copy.deepcopy(self.functions),
            domain=self.domain.copy(),
            density=self.density.copy(),
        )
        new_model.update()

        return new_model

    ## Model information
    # -------------------------
    def string_rep(self):
        l = [
            "model: {}".format(self.name),
            "",
            "  inputs:",
            "    var_det:",
            "".join([
                "      {}\n".format(self.domain.bound_summary(var_det))
                for var_det in self.var_det
            ]),
            "    var_rand:",
        ]

        try:
            l = l + [
                "".join([
                    "      {}\n".format(self.density.summary_marginal(var_rand))
                    for var_rand in self.density.marginals.keys()
                ]),
            ]
        except AttributeError:
            l = l + ["\n"]

        l = l + [
            "    copula:",
            "      {}\n".format(self.density.summary_copula()),
            "  functions:",
            "".join([
                "      {}\n".format(function.summary())
                for function in self.functions
            ])
        ]

        return "\n".join(l)

    def __str__(self):
        return self.string_rep()

    def __repr__(self):
        return self.string_rep()

    def printpretty(self):
        """Formatted print of model attributes
        """
        print(self.string_rep())

    def make_dag(self, expand=set()):
        """Generate a DAG for the model
        """
        G = nx.DiGraph()

        ## Inputs-to-Functions
        for f in self.functions:
            # Expand composed models
            if isinstance(f, FunctionModel) and (f.name in expand):
                G_ref = f.model.make_dag(expand=expand - {f})
                G_sub = nx.DiGraph()
                # Add nodes
                G_sub.add_node(f.name + ".var")
                G_sub.add_node(f.name + ".out")
                for g in f.model.functions:
                    G_sub.add_node(f.name + "." + g.name)
                # Add node metadata
                nx.set_node_attributes(G_sub, f.name, "parent")

                # Add edges
                for u, v, d in G_ref.edges(data=True):
                    # Add renamed edge
                    if u == "(var)":
                        G_sub.add_edge(f.name + ".var", f.name + "." + v, **d)
                    elif v == "(out)":
                        G_sub.add_edge(f.name + "." + u, f.name + ".out", **d)
                    else:
                        G_sub.add_edge(f.name + "." + u, f.name + "." + v, **d)

                # Compose the graphs
                G = nx.compose(G, G_sub)

            i_var = set(self.var).intersection(set(f.var))
            if len(i_var) > 0:
                s_var = "{}".format(i_var)
                if isinstance(f, FunctionModel) and (f.name in expand):
                    G.add_edge("(var)", f.name + ".var", label=s_var)
                else:
                    G.add_edge("(var)", f.name, label=s_var)

        ## Function-to-Function
        for i0 in range(len(self.functions)):
            for i1 in range(i0 + 1, len(self.functions)):
                f0 = self.functions[i0]
                f1 = self.functions[i1]
                i_var = set(f0.out).intersection(set(f1.var))

                ## If connected
                if len(i_var) > 0:
                    s_var = "{}".format(i_var)
                    ## Handle composed models
                    if isinstance(f0, FunctionModel) and (f0.name in expand):
                        name0 = f0.name + ".out"
                    else:
                        name0 = f0.name
                    if isinstance(f1, FunctionModel) and (f1.name in expand):
                        name1 = f1.name + ".out"
                    else:
                        name1 = f1.name

                    G.add_edge(name0, name1, label=s_var)

        ## Functions-to-Outputs
        for f in self.functions:
            i_out = set(self.out).intersection(set(f.out))

            if len(i_out) > 0:
                s_out = "{}".format(i_out)
                ## Target composed model's out
                if isinstance(f, FunctionModel) and (f.name in expand):
                    G.add_edge(f.name + ".out", "(out)", label=s_out)
                ## An ordinary function
                else:
                    G.add_edge(f.name, "(out)", label=s_out)

            # Add node metadata
            nx.set_node_attributes(G, {f.name: {"parent": self.name}})

        # Final metadata
        nx.set_node_attributes(G, {"(var)": {"parent": self.name}})
        nx.set_node_attributes(G, {"(out)": {"parent": self.name}})

        return G

    def show_dag(self, expand=set()):
        """Generate and show a DAG for the model
        """
        from matplotlib.pyplot import show as pltshow

        G = self.make_dag(expand=expand)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            ## Plotting
            edge_labels = dict(
                [((u, v,), d["label"]) for u, v, d in G.edges(data=True)]
            )
            n = G.size()

            ## Manual layout
            # if n == 2:
            if False:
                pos = {
                    "(var)": [-0.5, +0.5],
                    "(out)": [+0.5, -0.5],
                }
                pos[self.functions[0].name] = [+0.5, +0.5]
            ## Optimized layout
            else:
                try:
                    ## Planar, if possible
                    pos = nx.planar_layout(G)
                except nx.NetworkXException:
                    ## Scaled spring layout
                    pos = nx.spring_layout(
                        G,
                        k=0.6 * n,
                        pos={
                            "(Inputs)": [-0.5 * n, +0.5 * n],
                            "(Outputs)": [+0.5 * n, -0.5 * n],
                        },
                        fixed=["(var)", "(out)"],
                        threshold=1e-6,
                        iterations=100,
                    )

            # Generate colormap
            color_map = []
            for node in G:
                if G.nodes[node]["parent"] == self.name:
                    color_map.append("blue")
                else:
                    color_map.append("green")

            # Draw
            nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels)
            nx.draw(G, pos, node_size=1000, with_labels=True, node_color=color_map)
            pltshow()



__all__ = ["make_plate_buckle"]

from grama import cp_bounds, cp_copula_gaussian, cp_function, cp_marginals, \
    marg_named, Model
from grama.data import df_stang
from numpy import pi


LOAD = 0.00128  # Applied load (kips)

## Reference geometry
THICKNESS = 0.06  # Plate thickness (in)
HEIGHT = 12.0  # Plate height    (in)


def function_buckle_state(x):
    t, h, w, E, mu, L = x
    return pi * E / 12 / (1 - mu ** 2) * (t / h) ** 2 - L / t / w


def make_plate_buckle():
    md = (
        Model("Plate Buckling")
        >> cp_function(
            fun=function_buckle_state,
            var=["t", "h", "w", "E", "mu", "L"],
            out=["g_buckle"],
            name="limit state",
        )
        >> cp_bounds(
            t=(0.5 * THICKNESS, 2 * THICKNESS),
            h=(6, 18),
            w=(6, 18),
            L=(LOAD / 2, LOAD * 2),
        )
        >> cp_marginals(
            E=marg_named(df_stang.E, "norm"), mu=marg_named(df_stang.mu, "beta")
        )
        >> cp_copula_gaussian(df_data=df_stang)
    )

    return md



__all__ = ["make_trajectory_linear"]

from grama import cp_bounds, cp_vec_function, Model, df_make
from numpy import exp, Inf


## Constants
x0 = 0  # Initial x-position (m)
y0 = 0  # Initial y-position (m)
g = -9.8  # Gravitational acceleration (m/s^2)

## Responses (x and y trajectory components)
def fun_x(df):
    return df_make(
        x=df.tau * df.u0 * (1 - exp(-df.t / df.tau)) + x0
    )


def fun_y(df):
    v_inf = g * df.tau
    return df_make(
        y=df.tau * (df.v0 - v_inf) * (1 - exp(-df.t / df.tau)) + v_inf * df.t + y0
    )


# Units     (m/s) (m/s) (s)    (s)
var_list = ["u0", "v0", "tau", "t"]


def make_trajectory_linear():
    ## Assemble model
    md_trajectory = (
        Model("Trajectory Model")
        >> cp_vec_function(fun=fun_x, var=var_list, out=["x"], name="x_trajectory",)
        >> cp_vec_function(fun=fun_y, var=var_list, out=["y"], name="y_trajectory",)
        >> cp_bounds(
            u0=[0.1, Inf], v0=[0.1, Inf], tau=[0.05, Inf], t=[0, 600]
        )
    )

    return md_trajectory



__all__ = ["make_cantilever_beam"]

from grama import cp_bounds, cp_copula_independence, cp_function, cp_marginals, \
    Model
from collections import OrderedDict as od
from numpy import sqrt, array, Inf, float64
from scipy.stats import norm


LENGTH = 100
D_MAX  = 2.2535

MU_H   = 500.
MU_V   = 1000.
MU_E   = 2.9e7
MU_Y   = 40000.

TAU_H  = 100.
TAU_V  = 100.
TAU_E  = 1.45e6
TAU_Y  = 2000.

def function_area(x):
    w, t = x
    return w * t

def function_stress(x):
    w, t, H, V, E, Y = x
    return (Y - 600 * V / w / t**2 - 600 * H / w**2 / t) / MU_Y

def function_displacement(x):
    w, t, H, V, E, Y = x
    return D_MAX - float64(4) * LENGTH**3 / E / w / t * sqrt(
        V**2 / t**4 + H**2 / w**4
    )

def make_cantilever_beam():
    """Cantilever beam

    A standard reliability test-case, often used for benchmarking reliability
    analysis and design algorithms.

    Generally used in the following optimization problem:

        min_{w,t} c_area

        s.t.      P[g_stress <= 0] <= 1.35e-3

                  P[g_disp <= 0] <= 1.35e-3

                  1 <= w, t <= 4

    Deterministic Variables:
        w: Beam width
        t: Beam thickness
    Random Variables:
        H: Horizontal applied force
        V: Vertical applied force
        E: Elastic modulus
        Y: Yield stress
    Outputs:
        c_area: Cost; beam cross-sectional area
        g_stress: Limit state; stress
        g_disp: Limit state; tip displacement

    References:
        Wu, Y.-T., Shin, Y., Sues, R., and Cesare, M., "Safety-factor based approach for probability-based design optimization," American Institute of Aeronautics and Astronautics, Seattle, Washington, April 2001.
        Sues, R., Aminpour, M., and Shin, Y., "Reliability-based Multi-Disciplinary Optimiation for Aerospace Systems," American Institute of Aeronautics and Astronautics, Seattle, Washington, April 2001.

    """

    md = Model(name = "Cantilever Beam") >> \
         cp_function(
             fun=function_area,
             var=["w", "t"],
             out=["c_area"],
             name="cross-sectional area",
             runtime=1.717e-7
         ) >> \
         cp_function(
             fun=function_stress,
             var=["w", "t", "H", "V", "E", "Y"],
             out=["g_stress"],
             name="limit state: stress",
             runtime=8.88e-7
         ) >> \
         cp_function(
             fun=function_displacement,
             var=["w", "t", "H", "V", "E", "Y"],
             out=["g_disp"],
             name="limit state: displacement",
             runtime=3.97e-6
         ) >> \
         cp_bounds(
             w=(2, 4),
             t=(2, 4)
         ) >> \
         cp_marginals(
             H={"dist": "norm", "loc": MU_H, "scale": TAU_H, "sign": +1},
             V={"dist": "norm", "loc": MU_V, "scale": TAU_V, "sign": +1},
             E={"dist": "norm", "loc": MU_E, "scale": TAU_E, "sign":  0},
             Y={"dist": "norm", "loc": MU_Y, "scale": TAU_Y, "sign": -1}
         ) >> \
         cp_copula_independence()

    return md



__all__ = [
    "tran_feat_composition",
    "tf_feat_composition",
]

## Featurize with matminer
try:
    from matminer.featurizers.conversions import StrToComposition
    from matminer.featurizers.composition import ElementProperty

except ModuleNotFoundError:
    pass

from grama import add_pipe, pipe
from pandas import concat
from toolz import curry


## Compute matminer featurization
# --------------------------------------------------
@curry
def tran_feat_composition(
    df,
    var_formula="FORMULA",
    preset_name="magpie",
    append=True,
    ignore_errors=True,
    **kwargs,
):
    r"""Featurize a dataset using matminer

    Featurize chemical composition using matminer package.

    Args:
        df (DataFrame): Data to featurize
        var_formula (string): Column in df with chemical formula; formula
            given as string
        append (bool): Append results to original columns?
        preset_name (string): Matminer featurization preset

    Kwargs:
        ignore_errors (bool): Do not throw an error while parsing formulae; set to
            True to return NaN's for invalid formulae.

    Notes:
        - A pre-processor and wrapper for matminer.featurizers.composition

    References:
        Ward, L., Dunn, A., Faghaninia, A., Zimmermann, N. E. R., Bajaj, S., Wang, Q., Montoya, J. H., Chen, J., Bystrom, K., Dylla, M., Chard, K., Asta, M., Persson, K., Snyder, G. J., Foster, I., Jain, A., Matminer: An open source toolkit for materials data mining. Comput. Mater. Sci. 152, 60-69 (2018).

    Examples:
        >>> import grama as gr
        >>> from grama.tran import tf_feat_composition
        >>> (
        >>>     gr.df_make(FORMULA=["C6H12O6"])
        >>>     >> gr.tf_feat_composition()
        >>> )

    """
    ## Check invariants

    ## Featurize
    try:
        featurizer = ElementProperty.from_preset(preset_name=preset_name)

    except NameError as e:
        error_string = str(e)
        raise NameError(
            error_string +
            "\n\nThis function requires the `matminer` package. " +
            "Try running the following to install the package:\n"
            "    pip install matminer"
        )

    df_res = StrToComposition().featurize_dataframe(
        df[[var_formula]], var_formula, ignore_errors=ignore_errors,
    )
    df_res = featurizer.featurize_dataframe(
        df_res, col_id="composition", ignore_errors=ignore_errors, **kwargs,
    )
    df_res.drop(columns=[var_formula, "composition"], inplace=True)

    ## Concatenate as necessary
    if append:
        df_res = concat((df, df_res), axis=1)

    return df_res


tf_feat_composition = add_pipe(tran_feat_composition)



# Implementation of support points generator
#
# Reference:
#   Mak and Joseph, "Support Points" (2018) *The Annals of Statistics*

__all__ = [
    "tran_sp",
    "tf_sp",
]

from grama import add_pipe
from numpy import diag, eye, ma, newaxis, number, zeros
from numpy.linalg import norm
from numpy.random import choice, multivariate_normal
from numpy.random import seed as setseed
from pandas import DataFrame
from toolz import curry
from warnings import warn


## Helper functions
##################################################
def _iterate_x(X, Y, ind):
    r"""Iterate a single candidate point

    Implementation of Equation (22) from Mak and Joseph (2018)

    Arguments:
        X (np.array): candidate points, X.shape == (n, p)
        Y (np.array): target points, Y.shape == (N, p)
        ind (int): candidate to iterate, 0 <= ind <= n - 1

    Returns:
        np.array: updated candidate point
    """
    ## Setup
    n = X.shape[0]
    N = Y.shape[0]

    ## Compute iteration
    # First term
    diffx = ma.array(X[ind] - X, mask=False)
    diffx[ind].mask = True
    diffx_norm = ma.array(norm(diffx, axis=1), mask=False)
    diffx_norm.mask[ind] = True
    t1 = (N / n) * (diffx / diffx_norm[:, newaxis]).sum(axis=0)

    # Second term
    diffy_norm = norm(X[ind] - Y, axis=1)
    q = (1 / diffy_norm).sum()
    t2 = (Y / diffy_norm[:, newaxis]).sum(axis=0)

    return (1 / q) * (t1 + t2)


def _sp_cpp(X0, Y, delta=1e-6, iter_max=500):
    r"""Implementation of sp.cpp algorithm

    Implementation of sp.cpp algorithm from Mak and Joseph (2018). Note that
    this implementation takes

    Signature:
        X, d, iter_c = _sp_cpp(X0, Y)

    Arguments:
        X0 (np.array): initial candidate points, X0.shape == (n, p)
        Y (np.array): target points, Y.shape == (N, p)
        delta (float): convergence criterion, as average pairwise-distance
            between iterations
        iter_max (int): maximum iteration count

    Returns:
        X (np.array): optimized support points
        d (float): average pairwise-distance at termination
        iter_c (int): iteration count at termination

    """
    ## Setup
    N, p = Y.shape
    n = X0.shape[0]
    Xn = X0.copy()

    ## Primary loop
    d = delta * 2
    iter_c = 0
    # Check convergence criterion
    while (d >= delta) and (iter_c < iter_max):
        # Update the candidate points
        for i in range(n):
            Xn[i] = _iterate_x(X0, Y, i)

        # Update loop variables
        d = norm(X0 - Xn, axis=1).mean()
        iter_c = iter_c + 1

        # Overwrite X0
        X0 = Xn.copy()

    ## DEBUG
    return Xn, d, iter_c


def _perturbed_choice(Y, n):
    r"""Choose a set of perturbed points

    Arguments:
        Y (np.array): target points, Y.shape == (N, p)

    Returns:
        np.array: perturbed points, shape == (n, p)

    """
    i0 = choice(Y.shape[0], size=n)
    # Add noise to initial proposal to avoid X-Y overlap;
    # random directions with fixed distance
    V_rand = multivariate_normal(zeros(Y.shape[1]), eye(Y.shape[1]), size=n)
    V_rand = V_rand / norm(V_rand, axis=1)[:, newaxis]
    X0 = Y[i0] + V_rand * Y.std(axis=0)

    return X0


## Public interfaces
##################################################
@curry
def tran_sp(
    df,
    n=None,
    var=None,
    n_maxiter=500,
    tol=1e-3,
    seed=None,
    verbose=True,
    standardize=True,
):
    r"""Compact a dataset with support points

    Arguments:
        df (DataFrame): dataset to compact
        n (int): number of samples for compacted dataset
        var (list of str): list of variables to compact, must all be numeric
        n_maxiter (int): maximum number of iterations for support point algorithm
        tol (float): convergence tolerance
        verbose (bool): print messages to the console?
        standardize (bool): standardize columns before running sp? (Restores after sp)

    Returns:
        DataFrame: dataset compacted with support points

    Examples:
        >>> import grama as gr
        >>> from grama.data import df_diamonds
        >>> df_sp = gr.tran_sp(df_diamonds, n=50, var=["price", "carat"])
    """
    ## Setup
    setseed(seed)
    # Handle input variables
    if var is None:
        # Select numeric columns only
        var = list(df.select_dtypes(include=[number]).columns)
        if verbose:
            print("tran_sp has selected var = {}".format(var))
    # Extract values
    Y = df[var].values
    if standardize:
        Y_mean = Y.mean(axis=0)
        Y_sd = Y.std(axis=0)
        Y = (Y - Y_mean) / Y_sd
    # Generate initial proposal points
    X0 = _perturbed_choice(Y, n)

    ## Run sp.ccp algorithm
    X, d, iter_c = _sp_cpp(X0, Y, delta=tol, iter_max=n_maxiter)
    if verbose:
        print(
            "tran_sp finished in {0:} iterations with distance criterion {1:4.3e}".format(
                iter_c, d
            )
        )
        if d > tol:
            warn(
                "Convergence tolerance not met; d = {0:4.3e} > tol = {1:4.3e}".format(
                    d, tol
                ),
                RuntimeWarning,
            )

    if standardize:
        X = X * Y_sd + Y_mean

    ## Package results
    return DataFrame(data=X, columns=var)


tf_sp = add_pipe(tran_sp)



from setuptools import setup, find_packages

setup(
    name="py_grama",
    author="Zachary del Rosario",
    author_email="zdelrosario@outlook.com",
    version="0.3.0",
    packages=[
        "grama",
        "grama.data",
        "grama.dfply",
        "grama.eval",
        "grama.fit",
        "grama.models",
        "grama.tran",
        "grama.psdr",
    ],
    package_data={"grama.data": ["*.csv"]},
    license="MIT",
    url="https://github.com/zdelrosario/py_grama",
    description="A grammar of model analysis",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "matplotlib",
        "numpy>=1.15",
        "pandas",
        "plotnine>=0.8.0",
        "scipy",
        "toolz",
        "networkx",
    ],
    extras_require={
        "lolo": ["lolopy"],
        "matminer": ["pymatgen==2021.2.8.1", "matminer"],
        "sklearn": ["scikit-learn"],
        "statsmodels": ["statsmodels"],
        "pyDOE": ["pyDOE"],
        "umap": ["umap-learn"],
        "psdr": ["psdr"],
        "dev": [
            "lolopy",
            "pymatgen==2021.2.8.1",
            "matminer",
            "scikit-learn",
            "statsmodels",
            "pyDOE",
            "umap-learn",
            "psdr",
        ],
    }
)



## Function timing machinery
import timeit
import numpy as np
from numpy import sqrt, array, Inf

LENGTH = 100
D_MAX  = 2.2535

def wrapper(func, *args, **kwargs):
    def wrapped():
        return func(*args, **kwargs)
    return wrapped

def function_area(x):
    w, t = x
    return w * t

def function_stress(x):
    w, t, H, V, E, Y = x
    return Y - 600 * V / w / t**2 - 600 * H / w**2 / t

def function_displacement(x):
    w, t, H, V, E, Y = x
    return D_MAX - np.float64(4) * LENGTH**3 / E / w / t * sqrt(
        V**2 / t**4 + H**2 / w**4
    )

wr_area         = wrapper(function_area, [1] * 2)
wr_stress       = wrapper(function_stress, [1] * 6)
wr_displacement = wrapper(function_displacement, [1] * 6)

num = int(1e6)
total_area         = timeit.timeit(wr_area, number=num)
total_stress       = timeit.timeit(wr_stress, number=num)
total_displacement = timeit.timeit(wr_displacement, number=num)

print("time_area         = {}".format(total_area / num))
print("time_stress       = {}".format(total_stress / num))
print("time_displacement = {}".format(total_displacement / num))



__all__ = [
    "tran_group_by",
    "tf_group_by",
    "tran_ungroup",
    "tf_ungroup",
]

from .base import symbolic_evaluation
from .. import add_pipe


@symbolic_evaluation(eval_as_label=True)
def tran_group_by(df, *args):
    df._grouped_by = list(args)
    return df

tf_group_by = add_pipe(tran_group_by)


def tran_ungroup(df):
    df._grouped_by = None
    return df

tf_ungroup = add_pipe(tran_ungroup)



__all__ = [
    "add_pipe",
    "copy_meta",
    "custom_formatwarning",
    "df_equal",
    "df_make",
    "df_grid",
    "param_dist",
    "pipe",
    "valid_dist",
    "tran_outer",
    "tf_outer",
]

import warnings
from functools import wraps
from inspect import signature
from numbers import Integral
from pandas import DataFrame, concat
from pandas.testing import assert_frame_equal
from scipy.stats import alpha, anglit, arcsine, argus, beta, betaprime, \
    bradford, burr, burr12, cauchy, chi, chi2, cosine, crystalball, dgamma, \
    dweibull, erlang, expon, exponnorm, exponweib, exponpow, f, fatiguelife, \
    fisk, foldcauchy, foldnorm, gaussian_kde, genlogistic, gennorm, genpareto, \
    genexpon, genextreme, gausshyper, gamma, gengamma, genhalflogistic, \
    gilbrat, gompertz, gumbel_r, gumbel_l, halfcauchy, halflogistic, \
    halfnorm, halfgennorm, hypsecant, invgamma, invgauss, invweibull, \
    johnsonsb, johnsonsu, kappa4, kappa3, ksone, kstwobign, laplace, levy, \
    levy_l, levy_stable, logistic, loggamma, loglaplace, lognorm, lomax, \
    maxwell, mielke, moyal, nakagami, ncx2, ncf, nct, norm, norminvgauss, \
    pareto, pearson3, powerlaw, powerlognorm, powernorm, rdist, rayleigh, \
    rice, recipinvgauss, skewnorm, t, trapz, triang, truncexpon, truncnorm, \
    tukeylambda, uniform, vonmises, vonmises_line, wald, weibull_min, \
    weibull_max, wrapcauchy
from toolz import curry


## Scipy metadata
valid_dist = {
    "alpha": alpha,
    "anglit": anglit,
    "arcsine": arcsine,
    "argus": argus,
    "beta": beta,
    "betaprime": betaprime,
    "bradford": bradford,
    "burr": burr,
    "burr12": burr12,
    "cauchy": cauchy,
    "chi": chi,
    "chi2": chi2,
    "cosine": cosine,
    "crystalball": crystalball,
    "dgamma": dgamma,
    "dweibull": dweibull,
    "erlang": erlang,
    "expon": expon,
    "exponnorm": exponnorm,
    "exponweib": exponweib,
    "exponpow": exponpow,
    "f": f,
    "fatiguelife": fatiguelife,
    "fisk": fisk,
    "foldcauchy": foldcauchy,
    "foldnorm": foldnorm,
    # "frechet_r": frechet_r,
    # "frechet_l": frechet_l,
    "genlogistic": genlogistic,
    "gennorm": gennorm,
    "genpareto": genpareto,
    "genexpon": genexpon,
    "genextreme": genextreme,
    "gausshyper": gausshyper,
    "gamma": gamma,
    "gengamma": gengamma,
    "genhalflogistic": genhalflogistic,
    # "geninvgauss": geninvgauss,
    "gilbrat": gilbrat,
    "gompertz": gompertz,
    "gumbel_r": gumbel_r,
    "gumbel_l": gumbel_l,
    "halfcauchy": halfcauchy,
    "halflogistic": halflogistic,
    "halfnorm": halfnorm,
    "halfgennorm": halfgennorm,
    "hypsecant": hypsecant,
    "invgamma": invgamma,
    "invgauss": invgauss,
    "invweibull": invweibull,
    "johnsonsb": johnsonsb,
    "johnsonsu": johnsonsu,
    "kappa4": kappa4,
    "kappa3": kappa3,
    "ksone": ksone,
    "kstwobign": kstwobign,
    "laplace": laplace,
    "levy": levy,
    "levy_l": levy_l,
    "levy_stable": levy_stable,
    "logistic": logistic,
    "loggamma": loggamma,
    "loglaplace": loglaplace,
    "lognorm": lognorm,
    # "loguniform": loguniform,
    "lomax": lomax,
    "maxwell": maxwell,
    "mielke": mielke,
    "moyal": moyal,
    "nakagami": nakagami,
    "ncx2": ncx2,
    "ncf": ncf,
    "nct": nct,
    "norm": norm,
    "norminvgauss": norminvgauss,
    "pareto": pareto,
    "pearson3": pearson3,
    "powerlaw": powerlaw,
    "powerlognorm": powerlognorm,
    "powernorm": powernorm,
    "rdist": rdist,
    "rayleigh": rayleigh,
    "rice": rice,
    "recipinvgauss": recipinvgauss,
    "skewnorm": skewnorm,
    "t": t,
    "trapz": trapz,
    "triang": triang,
    "truncexpon": truncexpon,
    "truncnorm": truncnorm,
    "tukeylambda": tukeylambda,
    "uniform": uniform,
    "vonmises": vonmises,
    "vonmises_line": vonmises_line,
    "wald": wald,
    "weibull_min": weibull_min,
    "weibull_max": weibull_max,
    "wrapcauchy": wrapcauchy,
}

param_dist = {
    "alpha": ["a", "loc", "scale"],
    "anglit": ["loc", "scale"],
    "arcsine": ["loc", "scale"],
    "argus": ["chi", "loc", "scale"],
    "beta": ["a", "b", "loc", "scale"],
    "betaprime": ["a", "b", "loc", "scale"],
    "bradford": ["c", "loc", "scale"],
    "burr": ["c", "d", "loc", "scale"],
    "burr12": ["c", "d", "loc", "scale"],
    "cauchy": ["loc", "scale"],
    "chi": ["df", "loc", "scale"],
    "chi2": ["df", "loc", "scale"],
    "cosine": ["loc", "scale"],
    "crystalball": ["beta", "m", "loc", "scale"],
    "dgamma": ["a", "loc", "scale"],
    "dweibull": ["c", "loc", "scale"],
    "erlang": ["loc", "scale"],
    "expon": ["loc", "scale"],
    "exponnorm": ["K", "loc", "scale"],
    "exponweib": ["a", "c", "loc", "scale"],
    "exponpow": ["b", "loc", "scale"],
    "f": ["dfn", "dfd", "loc", "scale"],
    "fatiguelife": ["c", "loc", "scale"],
    "fisk": ["c", "loc", "scale"],
    "foldcauchy": ["c", "loc", "scale"],
    "foldnorm": ["c", "loc", "scale"],
    "frechet_r": ["c", "loc", "scale"],
    "frechet_l": ["c", "loc", "scale"],
    "genlogistic": ["c", "loc", "scale"],
    "gennorm": ["beta", "loc", "scale"],
    "genpareto": ["c", "loc", "scale"],
    "genexpon": ["a", "b", "c", "loc", "scale"],
    "genextreme": ["c", "loc", "scale"],
    "gausshyper": ["a", "b", "c", "z", "loc", "scale"],
    "gamma": ["a", "loc", "scale"],
    "gengamma": ["a", "c", "loc", "scale"],
    "genhalflogistic": ["c", "loc", "scale"],
    "geninvgauss": ["p", "b", "loc", "scale"],
    "gilbrat": ["loc", "scale"],
    "gompertz": ["c", "loc", "scale"],
    "gumbel_r": ["loc", "scale"],
    "gumbel_l": ["loc", "scale"],
    "halfcauchy": ["loc", "scale"],
    "halflogistic": ["loc", "scale"],
    "halfnorm": ["loc", "scale"],
    "halfgennorm": ["beta", "loc", "scale"],
    "hypsecant": ["loc", "scale"],
    "invgamma": ["a", "loc", "scale"],
    "invgauss": ["mu", "loc", "scale"],
    "invweibull": ["c", "loc", "scale"],
    "johnsonsb": ["a", "b", "loc", "scale"],
    "johnsonsu": ["a", "b", "loc", "scale"],
    "kappa4": ["h", "k", "loc", "scale"],
    "kappa3": ["a", "loc", "scale"],
    "ksone": ["n", "loc", "scale"],
    "kstwobign": ["loc", "scale"],
    "laplace": ["loc", "scale"],
    "levy": ["loc", "scale"],
    "levy_l": ["loc", "scale"],
    "levy_stable": ["alpha", "beta", "loc", "scale"],
    "logistic": ["loc", "scale"],
    "loggamma": ["c", "loc", "scale"],
    "loglaplace": ["c", "loc", "scale"],
    "lognorm": ["s", "loc", "scale"],
    "loguniform": ["a", "b", "loc", "scale"],
    "lomax": ["c", "loc", "scale"],
    "maxwell": ["loc", "scale"],
    "mielke": ["k", "s", "loc", "scale"],
    "moyal": ["loc", "scale"],
    "nakagami": ["nu", "loc", "scale"],
    "ncx2": ["df", "nc", "loc", "scale"],
    "ncf": ["dfn", "dfd", "nc", "loc", "scale"],
    "nct": ["df", "nc", "loc", "scale"],
    "norm": ["loc", "scale"],
    "norminvgauss": ["a", "b", "loc", "scale"],
    "pareto": ["b", "loc", "scale"],
    "pearson3": ["skew", "loc", "scale"],
    "powerlaw": ["a", "loc", "scale"],
    "powerlognorm": ["c", "s", "loc", "scale"],
    "powernorm": ["c", "loc", "scale"],
    "rdist": ["c", "loc", "scale"],
    "rayleigh": ["loc", "scale"],
    "rice": ["b", "loc", "scale"],
    "recipinvgauss": ["mu", "loc", "scale"],
    "semicircular": ["loc", "scale"],
    "skewnorm": ["a", "loc", "scale"],
    "t": ["df", "loc", "scale"],
    "trapz": ["c", "d", "loc", "scale"],
    "triang": ["c", "loc", "scale"],
    "truncexpon": ["b", "loc", "scale"],
    "truncnorm": ["a", "b", "loc", "scale"],
    "tukeylambda": ["lam", "loc", "scale"],
    "uniform": ["loc", "scale"],
    "vonmises": ["kappa", "loc", "scale"],
    "vonmises_line": ["kappa", "loc", "scale"],
    "wald": ["loc", "scale"],
    "weibull_min": ["c", "loc", "scale"],
    "weibull_max": ["c", "loc", "scale"],
    "wrapcauchy": ["c", "loc", "scale"],
}


## Core helper functions
##################################################
## Metadata copy
def copy_meta(df_source, df_target):
    """Internal metadata copy tool

    Args:
        df_source (DataFrame): Original dataframe
        df_target (DataFrame): Target dataframe; receives metadata

    Returns:
        DataFrame: df_target with copied metadata
    """
    df_target._grouped_by = getattr(df_source, "_grouped_by", None)
    df_target._plot_info = getattr(df_source, "_plot_info", None)
    df_target._meta = getattr(df_source, "_meta", None)

    return df_target


## Pipe decorator
class pipe(object):
    __name__ = "pipe"

    def __init__(self, function):
        self.function = function
        # self.__doc__ = function.__doc__
        self.chained_pipes = []

    def __rshift__(self, other):
        assert isinstance(other, pipe)
        self.chained_pipes.append(other)
        return self

    def __rrshift__(self, other):
        other_copy = other.copy()

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if isinstance(other, DataFrame):
                other_copy = copy_meta(other, other_copy)

        result = self.function(other_copy)

        for p in self.chained_pipes:
            result = p.__rrshift__(result)
        return result

    def __repr__(self):
        return (
            self.__name__
            + ": "
            + self.function.__name__
            + "\n Did you mean to use a full-prefix verb?"
        )

    def __call__(self, *args, **kwargs):
        return pipe(lambda x: self.function(x, *args, **kwargs))


## Pipe applicator
def add_pipe(fun):
    class NewPipe(pipe):
        __name__ = fun.__name__
        __doc__ = (
            fun.__doc__
            #"Pipe-enabled version of {}\n".format(fun)
            #+ "Inherited Signature: {}\n".format(signature(fun))
            #+ fun.__doc__
        )

    return NewPipe(fun)


## Safe length-checker
def safelen(x):
    try:
        if isinstance(x, str):
            raise TypeError
        return len(x)

    except TypeError:
        return 1


## DataFrame constructor utility
def df_make(**kwargs):
    r"""Construct a DataFrame

    Helper function to construct a DataFrame.

    Keyword Args:
        varname (iterable): Column for constructed dataframe; column
                            name inferred from variable name.
    Returns:
        DataFrame: Constructed DataFrame

    Preconditions:
        All provided iterables must have identical length or be of
        length one.

        All provided variable names (keyword arguments) must be distinct.

    Examples:
        A common use-case is to use df_make() to pass values to
        the df_det keyword argument succinctly;

        >>> import grama as gr
        >>> from models import make_test
        >>> md = make_test()
        >>> md >> \
        >>>     gr.ev_monte_carlo(
        >>>         n=1e3,
        >>>         df_det=gr.df_make(x2=[1, 2])
        >>>     )

    """
    ## Check lengths
    lengths = [safelen(v) for v in kwargs.values()]
    length_max = max(lengths)

    if not all([(l == length_max) | (l == 1) for l in lengths]):
        raise ValueError("Column lengths must be identical or one.")

    ## Construct dataframe
    df_res = DataFrame()
    for key in kwargs.keys():
        try:
            if len(kwargs[key]) > 1:
                df_res[key] = kwargs[key]
            else:
                df_res[key] = [kwargs[key][0]] * length_max
        except TypeError:
            df_res[key] = [kwargs[key]] * length_max

    return df_res


## DataFrame equality checker
def df_equal(df1, df2, close=False, precision=3):
    """Check DataFrame equality

    Check that two dataframes have the same columns and values. Allow column
    order to differ.

    Args:
        df1 (DataFrame): Comparison input 1
        df2 (DataFrame): Comparison input 2

    Returns:
        bool: Result of comparison

    """

    if not set(df1.columns) == set(df2.columns):
        return False

    if close:
        try:
            assert_frame_equal(
                df1[df2.columns],
                df2,
                check_dtype=False,
                check_exact=False,
                check_less_precise=precision,
            )
            return True
        except:
            return False
    else:
        return df1[df2.columns].equals(df2)


## Monkey-patched warning fcn
def custom_formatwarning(msg, *args, **kwargs):
    # ignore everything except the message
    return "Warning: " + str(msg) + "\n"


## DataFrame outer product
# --------------------------------------------------
@curry
def tran_outer(df, df_outer):
    r"""Outer merge

    Perform an outer-merge on two dataframes.

    Args:
        df (DataFrame): Data to merge
        df_outer (DataFrame): Data to merge; outer

    Returns:
        DataFrame: Merged data

    Examples:
        >>> import grama as gr
        >>> import pandas as pd
        >>> df = pd.DataFrame(dict(x=[1,2]))
        >>> df_outer = pd.DataFrame(dict(y=[3,4]))
        >>> df_res = gr.tran_outer(df, df_outer)
        >>> df_res
        >>>    x  y
        >>> 0  1  3
        >>> 1  2  3
        >>> 2  1  4
        >>> 3  2  4

    """
    # Check invariants
    if (df.shape[0] == 0) and (df_outer.shape[0] == 0):
        raise ValueError("At least one of df and df_outer must be non-empty")
    # Handle single-empty cases
    if (df.shape[0] == 0):
        return df_outer
    if (df_outer.shape[0] == 0):
        return df

    n_rows = df.shape[0]
    list_df = []

    for ind in range(df_outer.shape[0]):
        df_rep = concat([df_outer.iloc[[ind]]] * n_rows, ignore_index=True)
        list_df.append(concat((df, df_rep), axis=1))

    return concat(list_df, ignore_index=True)


tf_outer = add_pipe(tran_outer)


## DataFrame constructor utility; outer product
def df_grid(**kwargs):
    r"""Construct a DataFrame as outer-product

    Helper function to construct a DataFrame as an outer-product of the given
    columns.

    Keyword Args:
        varname (iterable): Column for constructed dataframe; column
                            name inferred from variable name.
    Returns:
        DataFrame: Constructed DataFrame

    Preconditions:
        All provided variable names (keyword arguments) must be distinct.

    Examples:
        A common use-case is to use df_grid() to define a sweep across variables.


    """
    ## Construct dataframe
    df_res = DataFrame()
    for key in kwargs.keys():
        # Switch based on argument length
        l = safelen(kwargs[key])
        if l == 1:
            df_tmp = DataFrame(columns=[key], data=[kwargs[key]])
        else:
            df_tmp = DataFrame(columns=[key], data=kwargs[key])

        df_res = tran_outer(df_res, df_outer=df_tmp)

    return df_res



__all__ = [
    "tran_umap",
    "tf_umap",
]

## Transforms via umap-learn package
try:
    from umap import UMAP

except ModuleNotFoundError:
    pass

from grama import add_pipe
from pandas import concat, DataFrame
from toolz import curry


## Compute UMAP
# --------------------------------------------------
@curry
def tran_umap(
    df, var=None, out="xi", keep=True, append=False, n_dim=2, seed=None, **kwargs
):
    r"""UMAP dimension reduction of a dataset

    Apply the UMAP algorithm to reduce the dimensionality of a dataset.

    Args:
        df (DataFrame): Hybrid point results from gr.eval_hybrid()
        var (list or None): Variables in df on which to perform dimension reduction.
            Use None to compute with all variables.
        out (string): Name of reduced-dimensionality output; indexed from 0 .. n_dim-1
        keep (bool): Keep unused columns (outside `var`) in new DataFrame?
        append (bool): Append results to original columns?
        n_dim (int): Target dimensionality

    Kwargs:
        n_neighbors (int): A smaller value emphasizes local structure, larger value emphasizes global structure. Assumed number of nearest-neighbors in clusters. Coenen and Pearce claim this is the most important hyperparameter for UMAP. default=15
        min_dist (float): Minimum distance between mapped points. default=0.1
        metric (str or function): Metric used for distance computations. See url: https://umap-learn.readthedocs.io/en/latest/parameters.html#metric

    Notes:
        - A wrapper for umap.UMAP

    References:
        - McInnes, L, Healy, J, UMAP: Uniform Manifold Approximation and Projection for Dimension Reduction, ArXiv e-prints 1802.03426, 2018
        - Andy Coenen, Adam Pearce "Understanding UMAP" url: https://pair-code.github.io/understanding-umap/

    Examples:

    """
    ## Check invariants
    if var is None:
        var = list(df.columns).copy()
    else:
        var = list(var).copy()
        diff = set(var).difference(set(df.columns))
        if len(diff) > 0:
            raise ValueError(
                "`var` must be subset of `df.columns`\n" "diff = {}".format(diff)
            )
    var_leftover = list(set(df.columns).difference(set(var)))

    ## Reduce dimensionality
    try:
        df_res = DataFrame(
            data=UMAP(n_components=n_dim, random_state=seed, **kwargs).fit_transform(
                df[var].values
            ),
            columns=[out + "{}".format(i) for i in range(n_dim)],
        )

    except NameError as e:
        error_string = str(e)
        raise NameError(
            error_string +
            "\n\nThis function requires the `umap` package. " +
            "Try running the following to install the package:\n"
            "    pip install umap-learn"
        )

    ## Concatenate as necessary
    if keep:
        df_res = concat(
            (df_res.reset_index(drop=True), df[var_leftover].reset_index(drop=True)),
            axis=1,
        )
    if append:
        df_res = concat(
            (df_res.reset_index(drop=True), df[var].reset_index(drop=True)), axis=1
        )

    return df_res


tf_umap = add_pipe(tran_umap)
